"""
helicopter monitoring and control module gas helicopters
"""

import os, sys, math, time

from pymavlink import mavutil
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_settings

class GasHeliModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(GasHeliModule, self).__init__(mpstate, "gas_heli", "Gas Heli", public=False)
        self.console.set_status('IGN', 'IGN', row=4)
        self.console.set_status('THR', 'THR', row=4)
        self.console.set_status('RPM', 'RPM: 0', row=4)
        self.add_command('gasheli', self.cmd_gasheli,
                         'gas helicopter control',
                         ['<start|stop>',
                          'set (GASHELISETTINGS)'])
        self.gasheli_settings = mp_settings.MPSettings(
            [ ('ignition_chan', int, 0),
              ('ignition_disable_time', float, 0.5),
              ('ignition_stop_time', float, 3),
              ('starter_chan', int, 0),
              ('starter_time', float, 3.0),
              ('starter_pwm_on', int, 2000),
              ('starter_pwm_off', int, 1000),
              ]
            )
        self.add_completion_function('(GASHELISETTINGS)', self.gasheli_settings.completion)
        self.starting_motor = False
        self.stopping_motor = False
        self.motor_t1 = None
        self.old_override = 0

    def mavlink_packet(self, msg):
        '''handle an incoming mavlink packet'''
        type = msg.get_type()

        master = self.master

        # add some status fields
        if type in [ 'RC_CHANNELS_RAW' ]:
            rc6 = msg.chan6_raw
            if rc6 > 1500:
                ign_colour = 'green'
            else:
                ign_colour = 'red'
            self.console.set_status('IGN', 'IGN', fg=ign_colour, row=4)

        if type in [ 'SERVO_OUTPUT_RAW' ]:
            rc8 = msg.servo8_raw
            if rc8 < 1200:
                thr_colour = 'red'
            elif rc8 < 1300:
                thr_colour = 'orange'
            else:
                thr_colour = 'green'
            self.console.set_status('THR', 'THR', fg=thr_colour, row=4)

        if type in [ 'RPM' ]:
            rpm = msg.rpm1
            if rpm < 3000:
                rpm_colour = 'red'
            elif rpm < 10000:
                rpm_colour = 'orange'
            else:
                rpm_colour = 'green'
            self.console.set_status('RPM', 'RPM: %u' % rpm, fg=rpm_colour, row=4)

    def valid_starter_settings(self):
        '''check starter settings'''
        if self.gasheli_settings.ignition_chan <= 0 or self.gasheli_settings.ignition_chan > 8:
            print("Invalid ignition channel %d" % self.gasheli_settings.ignition_chan)
            return False
        if self.gasheli_settings.starter_chan <= 0 or self.gasheli_settings.starter_chan > 14:
            print("Invalid starter channel %d" % self.gasheli_settings.starter_chan)
            return False
        return True

    def idle_task(self):
        '''run periodic tasks'''
        if self.starting_motor:
            if self.gasheli_settings.ignition_disable_time > 0:
                elapsed = time.time() - self.motor_t1
                if elapsed >= self.gasheli_settings.ignition_disable_time:
                    self.module('rc').set_override_chan(self.gasheli_settings.ignition_chan-1, self.old_override)
                    self.starting_motor = False
        if self.stopping_motor:
            elapsed = time.time() - self.motor_t1
            if elapsed >= self.gasheli_settings.ignition_stop_time:
                # hand back control to RC
                self.module('rc').set_override_chan(self.gasheli_settings.ignition_chan-1, self.old_override)
                self.stopping_motor = False

    def start_motor(self):
        '''start motor'''
        if not self.valid_starter_settings():
            return
        self.motor_t1 = time.time()
        self.stopping_motor = False

        if self.gasheli_settings.ignition_disable_time > 0:
            self.old_override = self.module('rc').get_override_chan(self.gasheli_settings.ignition_chan-1)
            self.module('rc').set_override_chan(self.gasheli_settings.ignition_chan-1, 1000)
            self.starting_motor = True
        else:
            # nothing more to do
            self.starting_motor = False

        # setup starter run
        self.master.mav.command_long_send(self.target_system,
                                          self.target_component,
                                          mavutil.mavlink.MAV_CMD_DO_REPEAT_SERVO, 0,
                                          self.gasheli_settings.starter_chan,
                                          self.gasheli_settings.starter_pwm_on,
                                          1,
                                          self.gasheli_settings.starter_time*2,
                                          0, 0, 0)
        print("Starting motor")

    def stop_motor(self):
        '''stop motor'''
        if not self.valid_starter_settings():
            return
        self.motor_t1 = time.time()
        self.starting_motor = False
        self.stopping_motor = True
        self.old_override = self.module('rc').get_override_chan(self.gasheli_settings.ignition_chan-1)
        self.module('rc').set_override_chan(self.gasheli_settings.ignition_chan-1, 1000)
        print("Stopping motor")

    def cmd_gasheli(self, args):
        '''gas help commands'''
        usage = "Usage: gasheli <start|stop|set>"
        if len(args) < 1:
            print(usage)
            return
        if args[0] == "start":
            self.start_motor()
        elif args[0] == "stop":
            self.stop_motor()
        elif args[0] == "set":
            self.gasheli_settings.command(args[1:])
        else:
            print(usage)

def init(mpstate):
    '''initialise module'''
    return GasHeliModule(mpstate)



#!/usr/bin/env python

from math import *

class LowPassFilter2p:
    def __init__(self, sample_freq, cutoff_freq):
        self.cutoff_freq = cutoff_freq
        self.sample_freq = sample_freq
        self.a1 = 0.0
        self.a2 = 0.0
        self.b0 = 0.0
        self.b1 = 0.0
        self.b2 = 0.0
        self.delay_element_1 = None
        self.delay_element_2 = None
        self.set_cutoff_frequency(sample_freq, cutoff_freq)

    def set_cutoff_frequency(self, sample_freq, cutoff_freq):
        self.cutoff_freq = cutoff_freq
        if self.cutoff_freq <= 0.0:
            return
        fr = self.sample_freq/self.cutoff_freq
        ohm = tan(pi/fr)
        c = 1.0+2.0*cos(pi/4.0)*ohm + ohm*ohm
        self.b0 = ohm*ohm/c
        self.b1 = 2.0*self.b0
        self.b2 = self.b0
        self.a1 = 2.0*(ohm*ohm-1.0)/c
        self.a2 = (1.0-2.0*cos(pi/4.0)*ohm+ohm*ohm)/c

    def apply(self, sample):
        if self.delay_element_1 is None:
            self.delay_element_1 = sample
            self.delay_element_2 = sample
        delay_element_0 = sample - (self.delay_element_1 * self.a1) - (self.delay_element_2 * self.a2)
        output = (delay_element_0 * self.b0) + (self.delay_element_1 * self.b1) + (self.delay_element_2 * self.b2)
    
        self.delay_element_2 = self.delay_element_1
        self.delay_element_1 = delay_element_0

        return output

if __name__ == "__main__":
    from pymavlink.rotmat import Vector3
    filter = LowPassFilter2p(1000.0, 98.0)

    v2 = filter.apply(Vector3(1,2,3))
    print(v2)



'''
readline handling for mavproxy
'''

import sys, glob, os, platform
from future.builtins import input
import re
from pymavlink import mavutil

# some python distributions don't have readline, so handle that case
# with a try/except
if platform.system() == 'Darwin':
    import gnureadline as readline
elif platform.system() == 'Windows' and sys.version_info >= (3, 0):
    # Python 3.8 defaults to Proactor event loop, which doesn't work well
    if sys.version_info >= (3, 8):
        import asyncio
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    from prompt_toolkit import prompt, PromptSession
    from prompt_toolkit.completion import Completer, Completion
    from prompt_toolkit.shortcuts import CompleteStyle
    from prompt_toolkit.patch_stdout import patch_stdout
else:
    try:
        import readline
    except ImportError:
        import pyreadline as readline
    
rline_mpstate = None
#redisplay = None

class rline(object):
    '''async readline abstraction'''
    def __init__(self, prompt, mpstate):
        global rline_mpstate
        self.prompt = prompt
        rline_mpstate = mpstate
        
        # other modules can add their own completion functions
        mpstate.completion_functions = {
            '(FILENAME)' : complete_filename,
            '(PARAMETER)' : complete_parameter,
            '(VARIABLE)' : complete_variable,
            '(SETTING)' : rline_mpstate.settings.completion,
            '(COMMAND)' : complete_command,
            '(ALIAS)' : complete_alias,
            '(AVAILMODULES)' : complete_modules,
            '(LOADEDMODULES)' : complete_loadedmodules
            }

        if platform.system() == 'Windows' and sys.version_info >= (3, 0):
            # Create key bindings registry with a custom binding for the Tab key that
            # displays completions like GNU readline.
            self.session = PromptSession()
            mpstate.completor = MAVPromptCompleter()

    def set_prompt(self, prompt):
        if prompt != self.prompt:
            self.prompt = prompt
            if platform.system() != 'Windows' or sys.version_info < (3, 0):
                sys.stdout.write(prompt)
                self.redisplay()

    def add_history(self, line):
        '''add a line to the history'''
        if platform.system() == 'Windows' and sys.version_info >= (3, 0):
            self.session.history.append_string(line)
        else:
            readline.add_history(line)
        self.redisplay()

    def redisplay(self):
        '''redisplay prompt'''
        try:
            redisplay()
        except Exception as ex:
            pass
            
    def get_prompt(self):
        '''return the current prompt'''
        return self.prompt
            
    def input(self):
        ''' get user input'''
        ret = ""
        if platform.system() == 'Windows' and sys.version_info >= (3, 0):
            global rline_mpstate
            with patch_stdout():
                return self.session.prompt(self.get_prompt,completer=rline_mpstate.completor, complete_while_typing=False, complete_style=CompleteStyle.READLINE_LIKE, refresh_interval=0.5)
        else:
            return input(self.get_prompt())

if platform.system() == 'Windows' and sys.version_info >= (3, 0):
    # This is the prompt-toolkit setup for Windows
    class MAVPromptCompleter(Completer):
        '''Completion generator for commands'''
        def __init__(self, ):
            #global rline_mpstate
            #self.mpstate = rline_mpstate
            self.last_clist = None
            # other modules can add their own completion functions

        def get_completions(self, document, complete_event):
            '''yield all the possible completion strings'''
            global rline_mpstate
            text = document.text_before_cursor
            #cmd = text.split(' ')
            
            cmd = re.split(" +", text)
            final_completor = cmd[-1]

            if len(cmd) != 0 and cmd[0] in rline_mpstate.completions:
                # we have a completion rule for this command
                self.last_clist = complete_rules(rline_mpstate.completions[cmd[0]], cmd[1:])
            elif len(cmd) == 0 or len(cmd) == 1:
                # if on first part then complete on commands and aliases
                self.last_clist = complete_command(text) + complete_alias(text)
            else:
                # assume completion by filename
                self.last_clist = glob.glob(text+'*')
            ret = []
            for c in self.last_clist:
                if c.startswith(final_completor) or c.startswith(final_completor.upper()):
                    ret.append(c)
            if len(ret) == 0:
                # if we had no matches then try case insensitively
                final_completor = final_completor.lower()
                for c in self.last_clist:
                    if c.lower().startswith(final_completor):
                        ret.append(c)

            for c in ret:
                yield Completion(c, start_position=-len(final_completor))

def complete_alias(text):
    '''return list of aliases'''
    global rline_mpstate
    return list(rline_mpstate.aliases.keys())

def complete_command(text):
    '''return list of commands'''
    global rline_mpstate
    return list(rline_mpstate.command_map.keys())

def complete_loadedmodules(text):
    global rline_mpstate
    return [ m.name for (m,pm) in rline_mpstate.modules ]

def complete_modules(text):
    '''complete mavproxy module names'''
    import MAVProxy.modules, pkgutil
    modlist = [x[1] for x in pkgutil.iter_modules(MAVProxy.modules.__path__)]
    ret = []
    loaded = set(complete_loadedmodules(''))
    for m in modlist:
        if not m.startswith("mavproxy_"):
            continue
        name = m[9:]
        if not name in loaded:
            ret.append(name)
    return ret

def complete_filename(text):
    '''complete a filename'''

    #ensure directories have trailing slashes:
    list = glob.glob(text+'*')
    for idx, val in enumerate(list):
        if os.path.isdir(val):
            list[idx] = (val + os.path.sep)

    return list

def complete_parameter(text):
    '''complete a parameter'''
    global rline_mpstate
    return list(rline_mpstate.mav_param.keys())

def complete_variable(text):
    '''complete a MAVLink variable or expression'''
    global rline_mpstate
    if text == '':
        return list(rline_mpstate.status.msgs.keys())

    if text.endswith(":2"):
        suffix = ":2"
        text = text[:-2]
    else:
        suffix = ''

    m1 = re.match("^(.*?)([A-Z0-9][A-Z0-9_]*(\[[0-9A-Z_]+\])?)[.]([A-Za-z0-9_]*)$", text)
    if m1 is not None:
        prefix = m1.group(1)
        mtype = m1.group(2)
        instance = m1.group(3)
        fname = m1.group(4)
        if mtype in rline_mpstate.status.msgs:
            ret = []
            for f in rline_mpstate.status.msgs[mtype].get_fieldnames():
                if f.startswith(fname):
                    ret.append(prefix + mtype + '.' + f + suffix)
            return ret
        return []

    # handle partially open arrays, like 'NAMED_VALUE_FLOAT[A'
    m1 = re.match("^(.*?)([A-Z0-9][A-Z0-9_]*(\[[0-9A-Z_]+))$", text)
    if m1 is not None:
        prefix = m1.group(1)
        mtype = m1.group(2)
        ret = []
        for k in list(rline_mpstate.status.msgs.keys()):
            if k.startswith(mtype):
                ret.append(prefix + k + suffix)
        if len(ret):
            return ret

    m2 = re.match("^(.*?)([A-Z0-9][A-Z0-9_]*(\[[0-9A-Z_]+\])?)$", text)
    prefix = m2.group(1)
    mtype = m2.group(2)
    ret = []
    for k in list(rline_mpstate.status.msgs.keys()):
        if k.startswith(mtype):
            ret.append(prefix + k + suffix)
    if len(ret):
        return ret

    try:
        if mavutil.evaluate_expression(text, rline_mpstate.status.msgs) is not None:
            return [text+suffix]
    except Exception as ex:
        pass

    return []

def rule_expand(component, text):
    '''expand one rule component'''
    global rline_mpstate
    if component[0] == '<' and component[-1] == '>':
        return component[1:-1].split('|')
    if component in rline_mpstate.completion_functions:
        return rline_mpstate.completion_functions[component](text)
    return [component]

def rule_match(component, cmd):
    '''see if one rule component matches'''
    if component == cmd:
        return True
    expanded = rule_expand(component, cmd)
    if cmd in expanded:
        return True
    return False

def complete_rule(rule, cmd):
    '''complete using one rule'''
    global rline_mpstate
    rule_components = rule.split(' ')
    
    #  complete the empty string (e.g "graph <TAB><TAB>")
    if len(cmd) == 0:
        return rule_expand(rule_components[0], "")

    # check it matches so far
    for i in range(len(cmd)-1):
        if not rule_match(rule_components[i], cmd[i]):
            return []

    # expand the next rule component
    expanded = []
    if platform.system() == 'Windows' and sys.version_info >= (3, 0):
        if len(rule_components) >= len(cmd):
            expanded = rule_expand(rule_components[len(cmd)-1], cmd[-1])
    else:
        expanded = rule_expand(rule_components[len(cmd)-1], cmd[-1])
    return expanded


def complete_rules(rules, cmd):
    '''complete using a list of completion rules'''
    if not isinstance(rules, list):
        rules = [rules]
    ret = []
    for r in rules:
        ret += complete_rule(r, cmd)
    return ret


last_clist = None

def complete(text, state):
    '''completion routine for when user presses tab'''
    global last_clist
    global rline_mpstate
    if state != 0 and last_clist is not None:
        return last_clist[state]

    # split the command so far
    cmd = re.split(" +", readline.get_line_buffer())

    if len(cmd) != 0 and cmd[0] in rline_mpstate.completions:
        # we have a completion rule for this command
        last_clist = complete_rules(rline_mpstate.completions[cmd[0]], cmd[1:])
    elif len(cmd) == 0 or len(cmd) == 1:
        # if on first part then complete on commands and aliases
        last_clist = complete_command(text) + complete_alias(text)
    else:
        # assume completion by filename
        last_clist = glob.glob(text+'*')
    ret = []
    for c in last_clist:
        if c.startswith(text) or c.startswith(text.upper()):
            ret.append(c)
    if len(ret) == 0:
        # if we had no matches then try case insensitively
        text = text.lower()
        for c in last_clist:
            if c.lower().startswith(text):
                ret.append(c)
    ret.append(None)
    last_clist = ret
    return last_clist[state]

# Configure readline for linux/mac systems
if platform.system() != 'Windows' or sys.version_info < (3, 0):
    readline.set_completer_delims(' \t\n;')
    readline.parse_and_bind("tab: complete")
    readline.set_completer(complete)
    #redisplay = readline.redisplay

if __name__ == "__main__":
    from mp_settings import MPSettings, MPSetting

    class mystate(object):
        def __init__(self):
            self.settings = MPSettings(
            [ MPSetting('foo', int, 1, 'foo int', tab='Link', range=(0,4), increment=1),
              MPSetting('bar', float, 4, 'bar float', range=(-1,20), increment=1)])
            self.completions = {
                "script" : ["(FILENAME)"],
                "set"    : ["(SETTING)"]
                }
            self.command_map = {
                'script'  : (None,   'run a script of MAVProxy commands'),
                'set'     : (None,   'mavproxy settings'),
                }
            self.aliases = {}

    state = mystate()
    rl = rline("test> ", state)
    while True:
        line = raw_input(rl.prompt)
        print("Got: %s" % line)



#!/usr/bin/env python
'''mode command handling'''

import time, os
from pymavlink import mavutil

from MAVProxy.modules.lib import mp_module

class ModeModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(ModeModule, self).__init__(mpstate, "mode", public=True)
        self.add_command('mode', self.cmd_mode, "mode change", self.available_modes())
        self.add_command('guided', self.cmd_guided, "fly to a clicked location on map")

    def cmd_mode(self, args):
        '''set arbitrary mode'''
        mode_mapping = self.master.mode_mapping()
        if mode_mapping is None:
            print('No mode mapping available')
            return
        if len(args) != 1:
            print('Available modes: ', mode_mapping.keys())
            return
        if args[0].isdigit():
            modenum = int(args[0])
        else:
            mode = args[0].upper()
            if mode not in mode_mapping:
                print('Unknown mode %s: ' % mode)
                return
            modenum = mode_mapping[mode]
        self.master.set_mode(modenum)

    def available_modes(self):
        if self.master is None:
            print('No mode mapping available')
            return []
        mode_mapping = self.master.mode_mapping()
        if mode_mapping is None:
            print('No mode mapping available')
            return []
        return mode_mapping.keys()

    def unknown_command(self, args):
        '''handle mode switch by mode name as command'''
        mode_mapping = self.master.mode_mapping()
        mode = args[0].upper()
        if mode in mode_mapping:
            self.master.set_mode(mode_mapping[mode])
            return True
        return False

    def cmd_guided(self, args):
        '''set GUIDED target'''
        if len(args) != 1 and len(args) != 3:
            print("Usage: guided ALTITUDE | guided LAT LON ALTITUDE")
            return

        if len(args) == 3:
            latitude = float(args[0])
            longitude = float(args[1])
            altitude = float(args[2])
            latlon = (latitude, longitude)
        else:
            latlon = self.mpstate.click_location
            if latlon is None:
                print("No map click position available")
                return
            altitude = float(args[0])

        print("Guided %s %s" % (str(latlon), str(altitude)))
        self.master.mav.mission_item_int_send (self.settings.target_system,
                                           self.settings.target_component,
                                           0,
                                           self.module('wp').get_default_frame(),
                                           mavutil.mavlink.MAV_CMD_NAV_WAYPOINT,
                                           2, 0, 0, 0, 0, 0,
                                           int(latlon[0]*1.0e7),
                                           int(latlon[1]*1.0e7),
                                           altitude)
                                           
    def mavlink_packet(self, m):
            mtype = m.get_type()
            if mtype == 'HIGH_LATENCY2':
                mode_map = mavutil.mode_mapping_bynumber(m.type)
                if mode_map and m.custom_mode in mode_map:
                    self.master.flightmode = mode_map[m.custom_mode]
            
            
def init(mpstate):
    '''initialise module'''
    return ModeModule(mpstate)



#!/usr/bin/env python
'''
param editor module
Akshath Singhal
June 2019
'''

import platform
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import multiproc
from MAVProxy.modules.mavproxy_paramedit import ph_event
import threading
from ..lib.wx_loader import wx
from MAVProxy.modules.mavproxy_paramedit import param_editor_frame
from pymavlink import mavutil
import time
ParamEditorEvent = ph_event.ParamEditorEvent


class ParamEditorEventThread(threading.Thread):
    def __init__(self, mp_paramedit, queue, lock):
        threading.Thread.__init__(self)
        self.mp_paramedit = mp_paramedit
        self.event_queue = queue
        self.event_queue_lock = lock
        self.time_to_quit = False
        self.queue_access_timeout = 0.5

    def module(self, name):
        '''access another module'''
        return self.mp_paramedit.mpstate.module(name)

    def master(self):
        '''access master mavlink connection'''
        return self.mp_paramedit.mpstate.master()

    def run(self):
        while not self.time_to_quit:
            while not self.event_queue.empty():
                try:
                    event = self.event_queue.get(block=False)
                    event_type = event.get_type()

                    if event_type == ph_event.PEE_READ_PARAM:
                        self.param_received = self.module('param').mav_param
                        self.mp_paramedit.gui_event_queue.put(ParamEditorEvent(
                            ph_event.PEGE_READ_PARAM, param=self.param_received, vehicle=self.mp_paramedit.mpstate.vehicle_name))

                    elif event_type == ph_event.PEE_TIME_TO_QUIT:
                        self.mp_paramedit.needs_unloading = True

                    elif event_type == ph_event.PEE_LOAD_FILE:
                        self.module('param').mav_param.load(
                                event.get_arg("path"))
                        self.param_received = self.module('param').mav_param
                        self.mp_paramedit.gui_event_queue.put(ParamEditorEvent(
                            ph_event.PEGE_READ_PARAM, param=self.param_received))

                    elif event_type == ph_event.PEE_SAVE_FILE:
                        self.module('param').mav_param.save(
                                event.get_arg("path"), verbose=True)

                    elif event_type == ph_event.PEE_WRITE_PARAM:
                        self.mp_paramedit.paramchanged = event.get_arg("modparam")
                        self.mp_paramedit.set_params()

                    elif event_type == ph_event.PEE_RESET:
                        master = self.mp_paramedit.mpstate.mav_master[0]
                        master.mav.command_long_send(master.target_system, master.target_component, mavutil.mavlink.MAV_CMD_PREFLIGHT_STORAGE, 0, 2.0, 0, 0, 0, 0, 0, 0)

                    elif event_type == ph_event.PEE_FETCH:
                        self.module('param').fetch_all()

                except Exception:
                    time.sleep(0.2)
            time.sleep(0.01)


class ParamEditorMain(object):
    def __init__(self, mpstate):
        self.param_received = {}
        self.paramchanged = {}
        self.fltmode_rc = None
        self.event_queue = multiproc.Queue()
        self.event_queue_lock = multiproc.Lock()
        self.gui_event_queue = multiproc.Queue()
        self.gui_event_queue_lock = multiproc.Lock()

        self.close_window = multiproc.Semaphore()
        self.close_window.acquire()

        self.mpstate = mpstate
        self.mpstate.param_editor = self
        self.needs_unloading = False

        if platform.system() == 'Windows':
            self.child = threading.Thread(
                            target=self.child_task,
                            args=(self.event_queue,
                                  self.event_queue_lock, self.gui_event_queue,
                                  self.gui_event_queue_lock, self.close_window))
        else:
            self.child = multiproc.Process(
                                target=self.child_task,
                                args=(self.event_queue,
                                      self.event_queue_lock, self.gui_event_queue,
                                      self.gui_event_queue_lock, self.close_window))

        self.child.start()

        self.event_thread = ParamEditorEventThread(
                            self, self.event_queue, self.event_queue_lock)
        self.event_thread.start()

        self.last_unload_check_time = time.time()
        self.unload_check_interval = 0.1  # seconds

        self.time_to_quit = False
        self.mavlink_message_queue = multiproc.Queue()
        self.mavlink_message_queue_handler = threading.Thread(
            target=self.mavlink_message_queue_handler)
        self.mavlink_message_queue_handler.start()

    def mavlink_message_queue_handler(self):
        while not self.time_to_quit:
            while True:
                if self.time_to_quit:
                    return
                if not self.mavlink_message_queue.empty():
                    break
                time.sleep(0.1)
            m = self.mavlink_message_queue.get()
            try:
                self.process_mavlink_packet(m)
            except Exception as ex:
                print(ex)
                import traceback
                traceback.print_stack()

    def unload(self):
        '''unload module'''
        self.mpstate.param_editor.close()

    def idle_task(self):
        now = time.time()
        if now - self.last_unload_check_time > self.unload_check_interval:
            self.last_unload_check_time = now
            if not self.child.is_alive():
                self.needs_unloading = True

    def mavlink_packet(self, m):
        if m.get_type() in ['PARAM_VALUE', 'RC_CHANNELS', 'RC_CHANNELS_RAW']:
            self.mavlink_message_queue.put(m)

    def process_mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        mtype = m.get_type()
        if mtype == 'PARAM_VALUE':
            if m.param_id in self.paramchanged:
                del self.paramchanged[m.param_id.upper()]
            self.gui_event_queue.put(ParamEditorEvent(
                ph_event.PEGE_WRITE_SUCC, paramid=m.param_id.upper(), paramvalue=m.param_value, pstatus = self.mpstate.module('param').param_status()))
        if mtype in ['RC_CHANNELS_RAW', 'RC_CHANNELS']:
            if self.mpstate.vehicle_name == 'APMrover2':
                fltmode_ch = int(self.mpstate.module('param').mav_param['MODE_CH'])
            else:
                if self.mpstate.vehicle_name.lower().find('copter') != -1:
                    default_channel = 5
                else:
                    default_channel = 8
                fltmode_ch = int(self.mpstate.module('param').mav_param.get('FLTMODE_CH', default_channel))
            if self.mpstate.vehicle_name is not None and fltmode_ch > 0:
                rc_received = float(getattr(m, 'chan%u_raw' % fltmode_ch))
                if rc_received != self.fltmode_rc and ((fltmode_ch > 0 and fltmode_ch < 9 and mtype == 'RC_CHANNELS_RAW') or (fltmode_ch > 0 and fltmode_ch < 19 and mtype == 'RC_CHANNELS')):
                    self.fltmode_rc = rc_received
                    self.gui_event_queue.put(ParamEditorEvent(
                        ph_event.PEGE_RCIN, rcin=rc_received))

    def child_task(self, queue, lock, gui_queue, gui_lock, close_window_sem):
        '''child process - this holds GUI elements'''
        mp_util.child_close_fds()
        self.app = wx.App(False)
        self.app.frame = param_editor_frame.ParamEditorFrame(
            parent=None, id=wx.ID_ANY)
        self.app.frame.set_event_queue(queue)
        self.app.frame.set_event_queue_lock(lock)
        self.app.frame.set_gui_event_queue(gui_queue)
        self.app.frame.set_gui_event_queue_lock(gui_lock)
        self.app.frame.get_vehicle_type(self.mpstate.vehicle_name)
        self.app.frame.set_close_window_semaphore(close_window_sem)
        self.app.frame.redirect_err(self.mpstate.settings.moddebug)
        self.app.frame.set_param_init(self.mpstate.module('param').mav_param, self.mpstate.vehicle_name)
        self.app.SetExitOnFrameDelete(True)
        self.app.frame.Show()

        # start a thread to monitor the "close window" semaphore:
        class CloseWindowSemaphoreWatcher(threading.Thread):
            def __init__(self, task, sem):
                threading.Thread.__init__(self)
                self.task = task
                self.sem = sem

            def run(self):
                self.sem.acquire(True)
                self.task.app.ExitMainLoop()
        watcher_thread = CloseWindowSemaphoreWatcher(self, close_window_sem)
        watcher_thread.start()

        self.app.MainLoop()
        # tell the watcher it is OK to quit:
        close_window_sem.release()
        watcher_thread.join()

    def close(self):
        '''close the Parameter Editor window'''
        self.time_to_quit = True
        self.close_window.release()
        if platform.system() == 'Windows':
            self.child.join()
        else:
            self.child.terminate()

    def set_params(self):
        for param, value in self.paramchanged.items():
            self.mpstate.mav_master[0].param_set_send(param, float(value))


def init(mpstate):
    '''initialise module'''
    return ParamEditorMain(mpstate)



#!/usr/bin/env python
import wx
import wx.grid as gridlib
import math


class GridCheckListEditor(gridlib.PyGridCellEditor):
    """
    This is a custom CheckListBox editor for setting of Bitmasks
    """
    def __init__(self, choice, pvalcol, start):
        # ensure the choices list has no gaps
        bvalue = 0
        self.choices = []
        for c in choice:
            bopt = c.split(":")
            if len(bopt) != 2:
                continue
            bval = int(bopt[0])
            while bvalue < bval:
                self.choices.append("%u:INVALID" % bvalue)
                bvalue += 1
            self.choices.append(c)
            bvalue += 1
        binary = bin(int(start))[2:]
        self.startValue = [(len(binary)-ones-1) for ones in range(len(binary)) if binary[ones] == '1']
        self.pvalcol = pvalcol
        gridlib.PyGridCellEditor.__init__(self)

    def set_checked(self, selected):
        if int(selected) < int(math.pow(2, len(self.choices))):
            binary = bin(int(selected))[2:]
            self.startValue = [(len(binary)-ones-1) for ones in range(len(binary)) if binary[ones] == '1']
            try:
                self._tc.SetChecked(self.startValue)
            except Exception as e:
                print (e)

    def get_checked(self):
        return self._tc.GetChecked()

    def Create(self, parent, id, evtHandler):
        self._tc = wx.CheckListBox(parent, id, choices=self.choices)
        self._tc.SetChecked(self.startValue)
        self.SetControl(self._tc)
        if evtHandler:
            self._tc.PushEventHandler(evtHandler)

    def SetSize(self, rect):
        self._tc.SetDimensions(rect.x, rect.y, rect.width+2, 30*len(self.choices),
                               wx.SIZE_ALLOW_MINUS_ONE)

    def Show(self, show, attr):
        super(GridCheckListEditor, self).Show(show, attr)

    def BeginEdit(self, row, col, grid):
        self._tc.SetChecked(self.startValue)

    def EndEdit(self, row, col, grid, oldVal):
        val = self._tc.GetChecked()
        self.startValue = val
        if val != oldVal:
            return self._tc.GetChecked()
        else:
            return None

    def get_pvalue(self):
        s = 0
        for i in self._tc.GetChecked():
            s = s + int(math.pow(2, i))
        return str(s)

    def ApplyEdit(self, row, col, grid):
        val = ""
        for i in self._tc.GetChecked():
            val = val + str(self.choices[i]) + "\n"
        grid.GetTable().SetValue(row, col, str(val))
        grid.GetTable().SetValue(row, self.pvalcol, self.get_pvalue())

    def Reset(self):
        self._tc.SetChecked(None)

    def StartingKey(self, evt):
        evt.Skip()

    def Destroy(self):
        super(GridCheckListEditor, self).Destroy()

    def Clone(self):
        return GridCheckListEditor(self.choices, self.pvalcol, self.startValue)


class GridDropListEditor(gridlib.PyGridCellEditor):
    """
    This is a custom DropBox editor for setting of Parameter Values
    """
    def __init__(self, choice, pvalcol, start=0):
        self.choices = choice
        self.startValue = int(start)
        self.pvalcol = pvalcol
        gridlib.PyGridCellEditor.__init__(self)

    def set_checked(self, selected):
        self.startValue = int(selected)
        try:
            self._tc.SetSelection(self.startValue)
        except Exception as e:
            print (e)

    def get_checked(self):
        return self._tc.GetSelection()

    def Create(self, parent, id, evtHandler):
        self._tc = wx.ComboBox(parent, id, choices=self.choices)
        self.set_checked(self.startValue)
        self.SetControl(self._tc)
        if evtHandler:
            self._tc.PushEventHandler(evtHandler)

    def Show(self, show, attr):
        super(GridDropListEditor, self).Show(show, attr)

    def SetSize(self, rect):
        self._tc.SetDimensions(rect.x, rect.y, rect.width+2, 30,
                               wx.SIZE_ALLOW_MINUS_ONE)

    def BeginEdit(self, row, col, grid):
        self._tc.SetSelection(self.startValue)

    def EndEdit(self, row, col, grid, oldVal):
        val = self._tc.GetSelection()
        if val != oldVal and val != -1:
            self.startValue = val
            return self._tc.GetSelection()
        else:
            return None

    def ApplyEdit(self, row, col, grid):
        grid.GetTable().SetValue(row, col, self._tc.GetStringSelection())
        grid.GetTable().SetValue(row, self.pvalcol, str(self._tc.GetStringSelection()).split(':')[0])

    def Reset(self):
        self._tc.SetSelection(None)

    def StartingKey(self, evt):
        evt.Skip()

    def Destroy(self):
        super(GridDropListEditor, self).Destroy()

    def Clone(self):
        return GridDropListEditor(self.choices, self.pvalcol, self.startValue)


class GridScrollEditor(gridlib.PyGridCellEditor):
    """
    This is a custom SpinControlDouble editor for setting of float values with given range and increments
    """
    def __init__(self, Range, pvalcol, start):
        self.Range = Range
        self.startValue = float(start)
        self.pvalcol = pvalcol
        gridlib.PyGridCellEditor.__init__(self)

    def set_checked(self, selected):
        self.startValue = selected
        try:
            self._tc.SetValue(selected)
        except Exception as e:
            print (e)

    def SetSize(self, rect):
        self._tc.SetDimensions(rect.x, rect.y, rect.width+2, 30,
                               wx.SIZE_ALLOW_MINUS_ONE)

    def get_checked(self):
        return self._tc.GetValue()

    def Create(self, parent, id, evtHandler):
        self._tc = wx.SpinCtrlDouble(parent, id, min=self.Range['Min'], max=self.Range['Max'], inc=self.Range['Increment'])
        self.SetControl(self._tc)
        self.set_checked(self.startValue)
        if evtHandler:
            self._tc.PushEventHandler(evtHandler)

    def Show(self, show, attr):
        super(GridScrollEditor, self).Show(show, attr)

    def BeginEdit(self, row, col, grid):
        self._tc.SetValue(self.startValue)

    def EndEdit(self, row, col, grid, oldVal):
        val = self._tc.GetValue()
        self.startValue = val
        if val != oldVal:
            return val
        else:
            return None

    def ApplyEdit(self, row, col, grid):
        grid.GetTable().SetValue(row, col, str(self._tc.GetValue()))
        grid.GetTable().SetValue(row, self.pvalcol, str(self._tc.GetValue()))

    def Reset(self):
        self._tc.SetValue(None)

    def StartingKey(self, evt):
        evt.Skip()

    def Destroy(self):
        super(GridScrollEditor, self).Destroy()

    def Clone(self):
        return GridScrollEditor(self.Range, self.pvalcol, self.startValue)



'''
multi-processing abstraction
This wraps the multiprocessing module, using billiard on MacOS
and multiprocessing on Linux and Windows

The key problem on MacOS is that you can't fork in any process that uses
threading, which is almost all of processes as so many libraries use
threads. So instead billiard uses an approach that uses fork+exec and re-runs
the script in the child. It is horrible, but it seems to be the only way to
make things work on MacOS
'''

class PipeQueue(object):
    '''simulate a queue using a pipe. This is used to avoid a problem with
    pipes on MacOS, while still keeping similar syntax'''
    def __init__(self):
        (self.sender,self.receiver) = Pipe()
        self.alive = True
        self.pending = []

    def close(self):
        self.alive = False
        self.sender.close()
        self.receiver.close()

    def put(self, *args):
        if not self.alive:
            return
        try:
            self.sender.send(*args)
        except Exception:
            self.close()

    def fill(self):
        if not self.alive:
            return
        try:
            while self.receiver.poll():
                m = self.receiver.recv()
                self.pending.append(m)
        except Exception:
            self.alive = False
            self.close()

    def get(self):
        if not self.alive:
            return None
        self.fill()
        if len(self.pending) > 0:
            return self.pending.pop(0)
        return None

    def qsize(self):
        self.fill()
        return len(self.pending)

    def empty(self):
        return self.qsize() == 0

import platform, os, sys

# we use billiard (and forking disable) on MacOS, and also if USE_BILLIARD environment
# is set. Using USE_BILLIARD allows for debugging of the crazy forking disable approach on
# a saner platform
# As of Python 3.8 the default start method for macOS is spawn and billiard is not required.
if ((platform.system() == 'Darwin' or os.environ.get('USE_BILLIARD',None) is not None)
    and sys.version_info < (3, 8)):
    from billiard import Process, forking_enable, freeze_support, Pipe, Semaphore, Event, Lock
    forking_enable(False)
    Queue = PipeQueue
else:
    from multiprocessing import Process, freeze_support, Pipe, Semaphore, Event, Lock, Queue



from paho.mqtt import MQTTException
from MAVProxy.modules.lib import mp_settings
from MAVProxy.modules.lib import mp_module
import paho.mqtt.client as mqtt
import json
import numbers


class MqttModule(mp_module.MPModule):

    def __init__(self, mpstate):
        super(MqttModule, self).__init__(mpstate, "mqtt", "mqtt publisher")
        self.client = mqtt.Client()
        self.device_prefix = ''
        self.mqtt_settings = mp_settings.MPSettings(
            [('ip', str, '127.0.0.1'),
             ('port', int, '1883'),
             ('name', str, 'mavproxy'),
             ('prefix', str, '')
             ])
        self.add_command('mqtt', self.mqtt_command, "mqtt module", ['connect', 'set (MQTTSETTING)'])
        self.add_completion_function('(MQTTSETTING)', self.mqtt_settings.completion)

    def mavlink_packet(self, m):
        """handle an incoming mavlink packet"""
        try:
            data = self.convert_to_dict(m)
            self.client.publish(f'{self.mqtt_settings.prefix}/{m.get_type()}', json.dumps(data))
        except MQTTException as e:
            print(f'mqtt: Exception occurred: {e}')

    def connect(self):
        """connect to mqtt broker"""
        try:
            self.client.reinitialise(client_id=self.mqtt_settings.name)
            print(f'connecting to {self.mqtt_settings.ip}:{self.mqtt_settings.port}')
            self.client.connect(self.mqtt_settings.ip, int(self.mqtt_settings.port), 30)
        except MQTTException as e:
            print(f'mqtt: could not establish connection: {e}')
            return
        print('connected...')

    def mqtt_command(self, args):
        """control behaviour of the module"""
        if len(args) == 0:
            print(self.usage())
        elif args[0] == 'set':
            self.mqtt_settings.command(args[1:])
        elif args[0] == 'connect':
            self.connect()

    def usage(self):
        """show help on command line options"""
        return "Usage: mqtt <set|connect>"

    def convert_to_dict(self, message):
        """converts mavlink message to python dict"""
        if hasattr(message, '_fieldnames'):
            result = {}
            for field in message._fieldnames:
                result[field] = self.convert_to_dict(getattr(message, field))
            return result
        if isinstance(message, numbers.Number):
            return message
        return str(message)


def init(mpstate):
    """initialise module"""
    return MqttModule(mpstate)



# Copyright (C) 2016  Intel Corporation. All rights reserved.
#
# This file is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
import math
import os
import time

from pymavlink.mavutil import mavlink
from wx.lib.wordwrap import wordwrap

from MAVProxy.modules import mavproxy_magical as magical
from MAVProxy.modules.lib import wavefront as wv
from MAVProxy.modules.lib.wx_loader import wx
from MAVProxy.modules.mavproxy_magical.wxgeodesicgrid import GeodesicGrid
from MAVProxy.modules.mavproxy_magical.wxvehicle import Vehicle

COMMON_FOREGROUND = '#212121'
LIGHT_BACKGROUND = '#fafafa'
DARK_BACKGROUND = '#d1d1d1'
PRIMARY_HIGHLIGHT = '#3359ec'
SECONDARY_HIGHLIGHT = '#0c666a'

def scale_color(color, scale):
    r, g, b = color
    r, g, b = round(scale * r), round(scale * g), round(scale * b)
    r, g, b = min(255, r), min(255, g), min(255, b)
    return wx.Colour(r, g, b)

# Class only for allowing getting a common default foreground color
class Panel(wx.Panel):
    def __init__(self, *k, **kw):
        super(Panel, self).__init__(*k, **kw)
        self.SetForegroundColour(COMMON_FOREGROUND)

# Button with a custom style
class Button(wx.PyControl):
    def __init__(self, *k, **kw):
        if 'style' not in kw:
            kw['style'] = wx.BORDER_NONE
        super(Button, self).__init__(*k, **kw)

        self.border_radius = 0
        self.border_width = 0
        self.border_color = None
        self.padding = wx.Size(0, 0)

        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_LEFT_UP, self.OnLeftUp)

        self.SetCursor(wx.StockCursor(wx.CURSOR_HAND))

    def OnLeftUp(self, evt):
        command_evt = wx.CommandEvent(wx.EVT_BUTTON.typeId, self.GetId())
        command_evt.SetEventObject(self)
        self.GetEventHandler().AddPendingEvent(command_evt);

    def GetBorderRadius(self):
        return self.border_radius

    def SetBorderRadius(self, radius):
        self.border_radius = radius

    def GetBorderWidth(self):
        return self.border_width

    def SetBorderWidth(self, width):
        self.border_width = width

    def GetBorderColor(self):
        if not self.border_color:
            return self.GetBackgroundColour()
        return self.border_color

    def SetBorderColor(self, color):
        self.border_color = color

    def GetPadding(self):
        return self.padding

    def SetPadding(self, size):
        self.padding = size

    def OnPaint(self, evt):
        dc = wx.BufferedPaintDC(self)
        self.Draw(dc)

    def Draw(self, dc):
        width, height = self.GetClientSize()
        if not width or not height:
            return

        gcdc = wx.GCDC(dc)

        gcdc.SetPen(wx.NullPen)

        bg = self.GetParent().GetBackgroundColour()
        gcdc.SetBackground(wx.Brush(bg, wx.SOLID))
        gcdc.Clear()

        border_radius = self.GetBorderRadius()
        border = self.GetBorderWidth()

        if border:
            gcdc.SetBrush(wx.Brush(self.GetBorderColor()))
            gcdc.DrawRoundedRectangle(0, 0, width, height, border_radius)

        w, h = width - 2 * border, height - 2 * border
        border_radius -= border

        gcdc.SetBrush(wx.Brush(self.GetBackgroundColour()))
        gcdc.DrawRoundedRectangle(border, border, w, h, border_radius)

        dc.SetFont(self.GetFont())
        label = self.GetLabel()
        w, h = dc.GetTextExtent(label)
        x, y = (width - w) / 2, (height - h) / 2
        dc.DrawText(label, x, y)

    def DoGetBestSize(self):
        dc = wx.ClientDC(self)
        dc.SetFont(self.GetFont())
        w, h = dc.GetTextExtent(self.GetLabel())
        pw, ph = self.GetPadding()
        w, h = w + pw, h + ph
        border = self.GetBorderWidth()
        w, h = w + 2 * border, h + 2 * border
        return wx.Size(w, h)

class CountdownText(wx.PyWindow):
    def __init__(self, *k, **kw):
        super(CountdownText, self).__init__(*k, **kw)

        self.value = 0
        self.border_color = COMMON_FOREGROUND

        self.Bind(wx.EVT_PAINT, self.OnPaint)

    def ShouldInheritColours(self):
        return True

    def GetBorderColor(self):
        return self.border_color

    def SetBorderColor(self, color):
        self.border_color = color

    def OnPaint(self, evt):
        dc = wx.BufferedPaintDC(self)
        self.Draw(dc)

    def Draw(self, dc):
        width, height = self.GetClientSize()
        if not width or not height:
            return

        gcdc = wx.GCDC(dc)

        bg = self.GetParent().GetBackgroundColour()
        gcdc.SetBackground(wx.Brush(bg, wx.SOLID))
        gcdc.Clear()

        x, y = width / 2, height / 2
        radius = min(x, y)
        border = int(radius / 20)
        gcdc.SetPen(wx.Pen(self.GetBorderColor(), border))
        gcdc.SetBrush(wx.Brush(bg, wx.SOLID))
        gcdc.DrawCircle(x, y, radius - border)

        # Set font size 70% of the diameter
        v = str(self.value)
        s = max(dc.GetTextExtent(v))
        scale = 2 * radius * .7 / s
        font = self.GetFont()
        font.SetPointSize(round(font.GetPointSize() * scale))
        dc.SetFont(font)

        w, h = dc.GetTextExtent(v)
        x, y = (width - w) / 2, (height - h) / 2
        dc.DrawText(v, x, y)

    def SetValue(self, value):
        self.value = value
        self.Refresh()

class ReportDialog(wx.Dialog):
    light_background = wx.Colour()
    light_background.SetFromString(LIGHT_BACKGROUND)
    light_background = scale_color(light_background, 1.2)

    class StatusIcon(wx.PyWindow):
        success_color = '#00ed00'
        failure_color = '#d81313'

        def __init__(self, *k, **kw):
            super(ReportDialog.StatusIcon, self).__init__(*k, **kw)
            self.success = True
            self.Bind(wx.EVT_PAINT, self.OnPaint)

        def Success(self, success):
            self.success = success
            self.Refresh()

        def OnPaint(self, evt):
            dc = wx.BufferedPaintDC(self)
            self.Draw(dc)

        def Draw(self, dc):
            width, height = self.GetClientSize()
            if not width or not height:
                return

            gcdc = wx.GCDC(dc)

            gcdc.SetPen(wx.NullPen)

            bg = self.GetParent().GetBackgroundColour()
            gcdc.SetBackground(wx.Brush(bg, wx.SOLID))
            gcdc.Clear()

            color = self.success_color if self.success else self.failure_color
            gcdc.SetBrush(wx.Brush(color))

            x = width / 2
            y = height / 2
            gcdc.DrawCircle(x, y, min(x, y))

    class CompassPanel(Panel):
        def __init__(self, parent, m, *k, **kw):
            super(ReportDialog.CompassPanel, self).__init__(parent, *k, **kw)

            self.InitUI()

            self.compass_text.SetLabel('Compass %d' % m.compass_id)

            if m.cal_status == mavlink.MAG_CAL_SUCCESS:
                self.status_icon.Success(True)
                self.status_text.SetLabel('SUCCESS')
            else:
                self.status_icon.Success(False)
                self.status_text.SetLabel('FAILURE')

            self.fitness_value_text.SetLabel('%.3f' % m.fitness)

            texts = (
                self.offsets_texts,
                self.diagonals_texts,
                self.offdiagonals_texts,
            )
            params = (
                (m.ofs_x, m.ofs_y, m.ofs_z),
                (m.diag_x, m.diag_y, m.diag_z),
                (m.offdiag_x, m.offdiag_y, m.offdiag_z),
            )
            for targets, values in zip(texts, params):
                for target, value in zip(targets, values):
                    target.SetLabel('%.2f' % value)

        def InitUI(self):
            color = wx.Colour()
            color.SetFromString(LIGHT_BACKGROUND)
            self.SetBackgroundColour(scale_color(color, 0.95))

            font = self.GetFont()
            font.SetWeight(wx.FONTWEIGHT_BOLD)
            font.SetPointSize(24)
            self.SetFont(font)

            self.SetMinSize((400, -1))

            self.compass_text = wx.StaticText(self)

            self.fitness_value_text = wx.StaticText(self)
            font.SetWeight(wx.FONTWEIGHT_NORMAL)
            self.fitness_value_text.SetFont(font)

            self.parameters_sizer = self.CalibrationParameters()

            fitness_sizer = wx.BoxSizer(wx.HORIZONTAL)
            text = wx.StaticText(self, label='Fitness')
            text.SetFont(self.fitness_value_text.GetFont())
            fitness_sizer.Add(text, proportion=1, flag=wx.EXPAND)
            fitness_sizer.Add(self.fitness_value_text)

            sizer = wx.BoxSizer(wx.VERTICAL)
            self.SetSizer(sizer)
            sizer.AddSpacer(16)
            sizer.Add(self.compass_text, border=16, flag=wx.LEFT | wx.RIGHT)
            sizer.AddSpacer(16)
            sizer.Add(self.StatusPanel(), flag=wx.EXPAND)
            sizer.AddSpacer(16)
            sizer.Add(fitness_sizer, border=16, flag=wx.LEFT | wx.RIGHT | wx.EXPAND)
            sizer.AddSpacer(16)
            sizer.Add(self.parameters_sizer, border=16, flag=wx.LEFT | wx.RIGHT | wx.BOTTOM | wx.EXPAND)

            sizer.Hide(self.parameters_sizer)

        def StatusPanel(self):
            panel = Panel(self)

            panel.SetBackgroundColour(ReportDialog.light_background)

            self.status_icon = ReportDialog.StatusIcon(panel, size=(44, 44))
            self.status_text = wx.StaticText(panel)

            sizer = wx.BoxSizer(wx.HORIZONTAL)
            panel.SetSizer(sizer)
            sizer.AddSpacer(16)
            sizer.Add(self.status_icon, border=5, flag=wx.TOP | wx.BOTTOM)
            sizer.AddSpacer(16)
            sizer.Add(self.status_text, flag=wx.ALIGN_CENTER)
            sizer.AddSpacer(16)

            return panel

        def CalibrationParameters(self):
            self.offsets_texts = tuple(wx.StaticText(self) for _ in range(3))
            self.diagonals_texts = tuple(wx.StaticText(self) for _ in range(3))
            self.offdiagonals_texts = tuple(wx.StaticText(self) for _ in range(3))

            table = (
                ('Offsets', self.offsets_texts),
                ('Diagonals', self.diagonals_texts),
                ('Off-diagonals', self.offdiagonals_texts),
            )

            sizer = wx.FlexGridSizer(len(table) + 1, 4, 4, 10)
            sizer.AddGrowableCol(0)

            font = self.GetFont()
            font.SetPointSize(14)
            font.MakeItalic()

            text = wx.StaticText(self, label='Parameter')
            text.SetFont(font)
            sizer.Add(text)
            for label in ('X', 'Y', 'Z'):
                text = wx.StaticText(self, label=label)
                text.SetFont(font)
                sizer.Add(text, flag=wx.ALIGN_CENTER)

            font.SetWeight(wx.FONTWEIGHT_NORMAL)

            for label, (x, y, z) in table:
                text = wx.StaticText(self)
                text.SetLabel(label)

                text.SetFont(font)
                x.SetFont(font)
                y.SetFont(font)
                z.SetFont(font)

                sizer.Add(text)
                sizer.Add(x, flag=wx.ALIGN_RIGHT)
                sizer.Add(y, flag=wx.ALIGN_RIGHT)
                sizer.Add(z, flag=wx.ALIGN_RIGHT)

            return sizer

        def ShowCompassParameters(self, show):
            sizer = self.GetSizer()
            if show:
                sizer.Show(self.parameters_sizer)
            else:
                sizer.Hide(self.parameters_sizer)

    def __init__(self, parent, mavlink_messages, *k, **kw):
        super(ReportDialog, self).__init__(parent, *k, **kw)

        self.compass_parameters_shown = False
        self.mavlink_messages = mavlink_messages

        for m in self.mavlink_messages:
            if m.cal_status != mavlink.MAG_CAL_SUCCESS:
                self.success = False
                break
        else:
            self.success = True

        self.InitUI()

    def InitUI(self):
        color = wx.Colour()
        color.SetFromString(DARK_BACKGROUND)
        self.SetBackgroundColour(scale_color(color, .9))

        cols = 2 if len(self.mavlink_messages) > 1 else 1
        compass_sizer = wx.GridSizer(0, cols, 2, 2)

        self.compass_panels = []
        for m in self.mavlink_messages:
            p = self.CompassPanel(self, m)
            compass_sizer.Add(p)
            self.compass_panels.append(p)

        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)
        sizer.Add(compass_sizer)
        sizer.Add(self.Footer(), flag=wx.EXPAND)
        sizer.Fit(self)

    def Footer(self):
        footer = Panel(self)

        footer.SetBackgroundColour(ReportDialog.light_background)

        font = footer.GetFont()
        font.SetPointSize(16)
        font.SetWeight(wx.FONTWEIGHT_BOLD)
        footer.SetFont(font)

        self.show_more_button = Button(footer)
        self.show_more_button.SetLabel('Show more')
        self.show_more_button.SetBackgroundColour(wx.TransparentColour)

        self.Bind(wx.EVT_BUTTON, self.OnShowMoreButton, self.show_more_button)

        if not self.success:
            self.try_again_button = Button(footer, wx.ID_REDO)
            self.try_again_button.SetLabel('TRY AGAIN')
            self.try_again_button.SetBackgroundColour(wx.TransparentColour)
            self.try_again_button.SetForegroundColour(PRIMARY_HIGHLIGHT)

            self.Bind(wx.EVT_BUTTON, self.OnActionButton, self.try_again_button)

        self.close_button = Button(footer, wx.ID_CLOSE)
        self.close_button.SetLabel('CLOSE')
        self.close_button.SetBackgroundColour(wx.TransparentColour)
        self.close_button.SetForegroundColour(PRIMARY_HIGHLIGHT)

        self.Bind(wx.EVT_BUTTON, self.OnActionButton, self.close_button)

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        footer.SetSizer(sizer)
        sizer.AddSpacer(16)
        sizer.Add(self.show_more_button)
        sizer.AddStretchSpacer()
        if not self.success:
            sizer.Add(self.try_again_button)
            sizer.AddSpacer(32)
        sizer.Add(self.close_button)
        sizer.AddSpacer(16)

        return footer

    def OnActionButton(self, evt):
        r = evt.GetEventObject().GetId()
        if self.IsModal():
            self.EndModal(r)
        else:
            self.SetReturnCode(r)
            self.Show(False)

    def OnShowMoreButton(self, evt):
        self.compass_parameters_shown = not self.compass_parameters_shown
        for p in self.compass_panels:
            p.ShowCompassParameters(self.compass_parameters_shown)

        if self.compass_parameters_shown:
            self.show_more_button.SetLabel('Show less')
        else:
            self.show_more_button.SetLabel('Show more')

        self.GetSizer().Fit(self)

# Unfortunately, wx.StaticText sucks at wrapping text automatically, thus we
# had to come up with our one (very simple) widget
class InstructionText(wx.PyWindow):
    def __init__(self, *k, **kw):
        super(InstructionText, self).__init__(*k, **kw)

        self.text = ''
        self.min_lines = 1

        self.Bind(wx.EVT_PAINT, self.OnPaint)

    def ShouldInheritColours(self):
        return True

    def SetText(self, text):
        self.text = text
        self.Refresh()

    def SetMinLines(self, lines):
        self.min_lines = lines

    def OnPaint(self, evt):
        dc = wx.BufferedPaintDC(self)
        self.Draw(dc)

    def Draw(self, dc):
        width, height = self.GetClientSize()
        if not width or not height:
            return

        bg = self.GetParent().GetBackgroundColour()
        dc.SetBackground(wx.Brush(bg, wx.SOLID))
        dc.Clear()

        dc.SetFont(self.GetFont())
        text = wordwrap(self.text, width, dc)
        lines = text.splitlines()
        _, h = dc.GetTextExtent(text)
        y = (height - h) / 2
        for line in lines:
            w, h = dc.GetTextExtent(line)
            x = (width - w) / 2
            dc.DrawText(line, x, y)
            y += h

    def DoGetBestSize(self):
        _, height = self.GetClientSize()
        dc = wx.ClientDC(self)
        dc.SetFont(self.GetFont())
        _, h = dc.GetTextExtent('\n'.join([' '] * self.min_lines))
        return -1, max(h, height)

class MagicalFrame(wx.Frame):
    initial_instruction = 'To start calibration, please place your vehicle as the image shows and press start.'
    countdown_instruction = 'Get ready!'
    calibration_instructions = (
        'Rotate the vehicle in as many different ways as possible.',
        'Some rotations can be repeated from different initial positions.',
        'You can take the rotations above as a reference.',
        'The image on the right is a visual feedback of the calibration progress.',
        'Don\'t worry about the rotation direction (clockwise or counterclockwise), both work.',
    )

    ref_rotations_script = ()

    for yaw in (0, math.pi / 2):
        ref_rotations_script += (
            ('.SetEuler', 0, 0, yaw),
            ('.SetAngvel', (1, 0, 0), math.pi / 2),
            ('wait', 4),
            ('.SetEuler', 0, 0, yaw),
            ('.SetAngvel', (0, 1, 0), math.pi / 2),
            ('wait', 4),
            ('.SetEuler', 0, 0, yaw),
            ('.SetAngvel', (1, 1, 0), math.pi / 2),
            ('wait', 4),
            ('.SetEuler', 0, 0, yaw),
            ('.SetAngvel', (1, -1, 0), math.pi / 2),
            ('wait', 4),
        )

    for (roll, pitch, yaw), rotation_axis in (
        ((0, 0, 0), 2),
        ((math.pi, 0, 0), 2),
        ((-math.pi / 2, 0, 0), 1),
        ((math.pi / 2, 0, 0), 1),
        ((0, -math.pi / 2, 0), 0),
        ((0, math.pi / 2, 0), 0),
    ):
        axis = [0, 0, 0]
        axis[rotation_axis] = 1
        ref_rotations_script += (
            ('.SetEuler', roll, pitch, yaw),
            ('.SetAngvel', axis, math.pi / 2),
            ('wait', 4),
        )


    ref_rotations_script += (('restart',),)

    def __init__(self, conn):
        super(MagicalFrame, self).__init__(None, title='Magical')

        self.state = None
        self.grid = None
        self.conn = conn
        self.button_state = None
        self.ui_is_active = False
        self.countdown = 0
        self.calibration_instruction_idx = 0

        self.InitUI()

        self.timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.OnTimer, self.timer)
        self.timer.Start(60)

        self.loader_timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.OnLoaderTimer, self.loader_timer)
        self.loader_timer.Start(50)

        self.countdown_timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.OnCountdownTimer, self.countdown_timer)

        self.calibration_instructions_timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.OnCalibrationInstructionsTimer,
                  self.calibration_instructions_timer)

        self.SetState('idle')
        self.main_panel.GetSizer().Fit(self)

        path = os.path.join(magical.datapath, 'quadcopter.obj')
        self.vehicle_loader = wv.ParserWorker(
            wv.ObjParser(filename=path),
            complete_callback=self.VehicleLoadCompleteCallback,
        )
        self.vehicle_loader.start()

    def VehicleLoadCompleteCallback(self):
        wx.CallAfter(self.VehicleLoadComplete)

    def VehicleLoadComplete(self):
        self.grid.SetVehicleWavefront(self.vehicle_loader.obj)
        self.ref_vehicle.SetVehicleWavefront(self.vehicle_loader.obj)
        self.vehicle_loader = None
        self.loader_timer.Stop()
        self.HideLoadProgress()

    def InitUI(self):
        font = self.GetFont()
        font.SetFamily(wx.FONTFAMILY_SWISS)
        self.SetFont(font)

        self.main_panel = wx.ScrolledWindow(self)
        self.main_panel.SetScrollbars(1, 1, 1, 1)
        self.main_panel.SetBackgroundColour(DARK_BACKGROUND)

        sizer = wx.BoxSizer(wx.VERTICAL)
        self.main_panel.SetSizer(sizer)

        kw = dict(border=32, flag=wx.EXPAND | wx.LEFT | wx.RIGHT)

        sizer.AddSpacer(16)
        sizer.Add(self.Header(self.main_panel), **kw)
        sizer.AddSpacer(23)
        sizer.Add(self.Body(self.main_panel), proportion=1, **kw)
        sizer.AddSpacer(23)
        sizer.Add(self.Footer(self.main_panel), **kw)
        sizer.AddSpacer(23)

    def Header(self, parent):
        header = Panel(parent)
        header.SetBackgroundColour(LIGHT_BACKGROUND)

        sizer = wx.BoxSizer(wx.VERTICAL)
        header.SetSizer(sizer)

        font = header.GetFont()
        font.SetWeight(wx.FONTWEIGHT_BOLD)

        text = wx.StaticText(header)
        text.SetLabel('Compass Calibration')
        text.SetForegroundColour(PRIMARY_HIGHLIGHT)
        font.SetPointSize(26)
        text.SetFont(font)
        sizer.Add(text, border=16, flag=wx.ALL & ~wx.BOTTOM)

        text = wx.StaticText(header)
        text.SetLabel('Follow the procedure to calibrate your compass')
        font.SetPointSize(14)
        text.SetFont(font)
        sizer.Add(text, border=16, flag=wx.ALL & ~wx.TOP)

        return header

    def Body(self, parent):
        body = wx.BoxSizer(wx.HORIZONTAL)
        body.Add(self.InstructionsPanel(parent), proportion=2, flag=wx.EXPAND)
        body.AddSpacer(23)
        self.progress_panel = self.ProgressPanel(parent)
        body.Add(self.progress_panel, proportion=1, flag=wx.EXPAND)
        return body

    def InstructionsPanel(self, parent):
        panel = Panel(parent)
        panel.SetBackgroundColour(LIGHT_BACKGROUND)

        self.ref_vehicle = Vehicle(panel)
        self.ref_vehicle.SetMinSize((430, 430))

        self.countdown_window = CountdownText(panel)
        self.countdown_window.SetMinSize((430, 430))
        self.countdown_window.SetBorderColor(SECONDARY_HIGHLIGHT)
        self.countdown_window.SetForegroundColour(PRIMARY_HIGHLIGHT)
        font = self.countdown_window.GetFont()
        font.SetWeight(wx.FONTWEIGHT_BOLD)
        self.countdown_window.SetFont(font)

        self.instruction = InstructionText(panel)
        self.instruction.SetMinLines(3)
        font = self.instruction.GetFont()
        font.SetPointSize(26)
        self.instruction.SetFont(font)

        sizer = wx.BoxSizer(wx.VERTICAL)
        panel.SetSizer(sizer)
        sizer.AddStretchSpacer()
        sizer.Add(self.ref_vehicle, flag=wx.ALIGN_CENTER)
        sizer.Add(self.countdown_window, flag=wx.ALIGN_CENTER)
        sizer.AddStretchSpacer()
        sizer.Add(self.instruction, border=10, flag=wx.ALL | wx.EXPAND)
        sizer.AddStretchSpacer()

        sizer.Hide(self.countdown_window)

        return panel

    def ProgressPanel(self, parent):
        panel = Panel(parent)
        panel.SetBackgroundColour(LIGHT_BACKGROUND)

        self.progress = wx.StaticText(panel, label='0%')
        font = self.progress.GetFont()
        font.SetPointSize(70)
        font.SetWeight(wx.FONTWEIGHT_BOLD)
        self.progress.SetFont(font)

        self.grid = GeodesicGrid(panel)
        self.grid.SetMinSize((365, 365))

        sizer = wx.BoxSizer(wx.VERTICAL)
        panel.SetSizer(sizer)
        sizer.AddStretchSpacer()
        sizer.Add(self.grid, flag=wx.ALIGN_CENTER)
        sizer.AddStretchSpacer()
        sizer.AddSpacer(16)
        sizer.Add(self.progress, flag=wx.ALIGN_CENTER)
        sizer.AddStretchSpacer()

        return panel

    def Footer(self, parent):
        footer = wx.BoxSizer(wx.HORIZONTAL)

        self.button = button = Button(parent, size=(180, -1))
        button.SetPadding((0, 16))
        button.SetBackgroundColour(LIGHT_BACKGROUND)
        button.SetBorderRadius(10)
        button.SetBorderWidth(2)
        font = button.GetFont()
        font.SetWeight(wx.FONTWEIGHT_BOLD)
        font.SetPointSize(20)
        button.SetFont(font)

        self.Bind(wx.EVT_BUTTON, self.OnButton, button)

        footer.Add(self.LoaderProgress(parent), proportion=1, flag=wx.EXPAND)
        footer.Add(button)

        return footer

    def LoaderProgress(self, parent):
        panel = wx.Panel(parent)
        panel.SetBackgroundColour(parent.GetBackgroundColour())

        font = panel.GetFont()
        font.MakeItalic()
        panel.SetFont(font)

        self.load_progress = wx.Gauge(panel, size=(-1, 10))
        self.load_progress_last_time = time.time()
        self.load_text = wx.StaticText(panel)

        sizer = wx.BoxSizer(wx.VERTICAL)
        panel.SetSizer(sizer)
        sizer.Add(self.load_progress, border=10, flag=wx.EXPAND | wx.LEFT | wx.RIGHT)
        sizer.Add(self.load_text, border=10, flag=wx.LEFT | wx.RIGHT)

        return panel

    def UpdateLoaderProgress(self, text, value, range):
        if value == -1:
            t = time.time()
            if t - self.load_progress_last_time > .001:
                self.load_progress.Pulse()
                self.load_progress_last_time = t
        else:
            self.load_progress.SetRange(range)
            self.load_progress.SetValue(value)
        if text != self.load_text.GetLabel():
            self.load_text.SetLabel(text)
        self.ShowLoadProgress()

    def ShowLoadProgress(self):
        sizer = self.load_progress.GetContainingSizer()
        sizer.Show(self.load_progress)
        sizer.Show(self.load_text)

    def HideLoadProgress(self):
        sizer = self.load_progress.GetContainingSizer()
        sizer.Hide(self.load_progress)
        sizer.Hide(self.load_text)

    def SetState(self, state):
        if self.state == state:
            return

        if self.state == 'countdown':
            self.countdown = 0
            self.countdown_timer.Stop()
            self.countdown_window.SetValue(self.countdown)

            sizer = self.countdown_window.GetContainingSizer()
            sizer.Show(self.ref_vehicle)
            sizer.Hide(self.countdown_window)
            sizer.Layout()

        elif self.state == 'running':
            self.calibration_instructions_timer.Stop()
            self.ref_vehicle.StopScript()
            self.ref_vehicle.SetEuler(0, 0, 0)

        if state == 'idle':
            self.SetInstructionText(self.initial_instruction)
            self.SetButtonState('start')

        elif state == 'countdown':
            self.countdown = 5
            self.countdown_window.SetValue(self.countdown)
            self.SetInstructionText(self.countdown_instruction)
            self.countdown_timer.Start(1000)
            self.SetButtonState('cancel')

            sizer = self.countdown_window.GetContainingSizer()
            sizer.Show(self.countdown_window)
            sizer.Hide(self.ref_vehicle)
            sizer.Layout()

        elif state == 'start':
            self.conn.send('start')

        elif state == 'running':
            self.calibration_instruction_idx = 0
            self.SetInstructionText(self.calibration_instructions[0])
            self.calibration_instructions_timer.Start(5000)

            self.SetButtonState('cancel')
            self.ref_vehicle.RunScript(self.ref_rotations_script)

        elif state == 'cancel':
            self.conn.send('cancel')

            if self.state == 'countdown':
                self.SetState('idle')

        self.state = state

    def OnButton(self, evt):
        if self.button_state == 'start':
            self.SetState('countdown')
        elif self.button_state == 'cancel':
            self.SetState('cancel')

    def SetInstructionText(self, text):
        self.instruction.SetText(text)
        self.instruction.GetContainingSizer().Layout()

    def OnCountdownTimer(self, evt):
        self.countdown -= 1
        self.countdown_window.SetValue(self.countdown)
        if not self.countdown:
            self.SetState('start')

    def OnLoaderTimer(self, evt):
        if self.vehicle_loader and self.vehicle_loader.is_alive():
            i, num_lines = self.vehicle_loader.get_progress()
            self.UpdateLoaderProgress('Loading 3D vehicle object...', i, num_lines)

    def OnTimer(self, evt):
        close_requested = False
        while self.conn.poll():
            msg = self.conn.recv()

            if msg['name'] == 'ui_is_active':
                self.ui_is_active = msg['value']
            elif msg['name'] == 'close':
                close_requested = True
            elif msg['name'] == 'running':
                if msg['value']:
                    self.SetState('running')
                else:
                    if self.state != 'countdown':
                        self.SetState('idle')
            elif msg['name'] == 'progress_update':
                self.UpdateProgress(msg['pct'], msg['sections'])
            elif msg['name'] == 'report':
                self.Report(msg['messages'], self.ui_is_active)
            elif msg['name'] == 'attitude':
                self.grid.SetAttitude(msg['roll'], msg['pitch'], msg['yaw'], msg['timestamp'])
            elif msg['name'] == 'mag':
                self.grid.SetMag(msg['x'], msg['y'], msg['z'])

        if close_requested:
            self.timer.Stop()
            self.Destroy()

    def OnCalibrationInstructionsTimer(self, evt):
        self.calibration_instruction_idx += 1
        self.calibration_instruction_idx %= len(self.calibration_instructions)
        text = self.calibration_instructions[self.calibration_instruction_idx]
        self.SetInstructionText(text)

    def UpdateProgress(self, pct, sections):
        self.progress.SetLabel("%d%%" % pct)
        self.progress.GetContainingSizer().Layout()
        if self.grid:
            self.grid.UpdateVisibleSections(sections)

    def SetButtonState(self, state):
        if state == self.button_state:
            return

        if state == 'start':
            self.button.SetLabel('START')
            self.button.SetForegroundColour(PRIMARY_HIGHLIGHT)
            self.button.SetBorderColor(PRIMARY_HIGHLIGHT)
        elif state == 'cancel':
            self.button.SetLabel('CANCEL')
            self.button.SetBorderColor(COMMON_FOREGROUND)
            self.button.SetForegroundColour(COMMON_FOREGROUND)

        self.button_state = state
        self.button.Refresh()

    def Report(self, mavlink_messages, popup):
        self.SetState('idle')
        for m in mavlink_messages:
            if m.cal_status != mavlink.MAG_CAL_SUCCESS:
                break
        else:
            self.UpdateProgress(100, [True] * 80)

        if popup:
            p = ReportDialog(self, mavlink_messages)
            r = p.ShowModal()
            if r == wx.ID_REDO:
                self.SetState('countdown')



import time, os
from pymavlink import mavutil, mavparm
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import multiproc

class ParamHelp:
    def __init__(self):
        self.xml_filepath = None
        self.vehicle_name = None
        self.last_pair = (None,None)
        self.last_htree = None

    def param_help_download(self):
        '''download XML files for parameters'''
        files = []
        for vehicle in ['Rover', 'ArduCopter', 'ArduPlane', 'ArduSub', 'AntennaTracker', 'Blimp']:
            url = 'http://autotest.ardupilot.org/Parameters/%s/apm.pdef.xml.gz' % vehicle
            path = mp_util.dot_mavproxy("%s.xml" % vehicle)
            files.append((url, path))
            url = 'http://autotest.ardupilot.org/%s-defaults.parm' % vehicle
            if vehicle != 'AntennaTracker':
                # defaults not generated for AntennaTracker ATM
                path = mp_util.dot_mavproxy("%s-defaults.parm" % vehicle)
                files.append((url, path))
        try:
            child = multiproc.Process(target=mp_util.download_files, args=(files,))
            child.start()
        except Exception as e:
            print(e)

    def param_use_xml_filepath(self, filepath):
        self.xml_filepath = filepath

    def param_help_tree(self, verbose=False):
        '''return a "help tree", a map between a parameter and its metadata.  May return None if help is not available'''
        if self.last_pair == (self.xml_filepath, self.vehicle_name):
            return self.last_htree
        if self.xml_filepath is not None:
            if verbose:
                print("param: using xml_filepath=%s" % self.xml_filepath)
            path = self.xml_filepath
        else:
            if self.vehicle_name is None:
                if verbose:
                    print("Unknown vehicle type")
                return None
            path = mp_util.dot_mavproxy("%s.xml" % self.vehicle_name)
            if not os.path.exists(path):
                if self.vehicle_name == 'APMrover2':
                    path = mp_util.dot_mavproxy("%s.xml" % "Rover")
            if not os.path.exists(path):
                if verbose:
                    print("Please run 'param download' first (vehicle_name=%s)" % self.vehicle_name)
                return None
        if not os.path.exists(path):
            if verbose:
                print("Param XML (%s) does not exist" % path)
            return None
        xml = open(path,'rb').read()
        from lxml import objectify
        objectify.enable_recursive_str()
        tree = objectify.fromstring(xml)
        htree = {}
        for p in tree.vehicles.parameters.param:
            n = p.get('name').split(':')[1]
            htree[n] = p
        for lib in tree.libraries.parameters:
            for p in lib.param:
                n = p.get('name')
                htree[n] = p
        self.last_htree = htree
        self.last_pair = (self.xml_filepath, self.vehicle_name)
        return htree

    def param_set_xml_filepath(self, args):
        self.xml_filepath = args[0]

    def param_apropos(self, args):
        '''search parameter help for a keyword, list those parameters'''
        if len(args) == 0:
            print("Usage: param apropos keyword")
            return

        htree = self.param_help_tree(True)
        if htree is None:
            return

        contains = {}
        for keyword in args:
            keyword = keyword.lower()
            for param in htree.keys():
                if str(htree[param]).lower().find(keyword) != -1:
                    contains[param] = True
        for param in contains.keys():
            print("%s" % (param,))

    def get_Values_from_help(self, help):
        children = help.getchildren()
        if len(children) == 0:
            return []
        vchild = children[0]
        return vchild.getchildren()

    def get_bitmask_from_help(self, help):
        if not hasattr(help, 'field'):
            return None
        field = help.field
        if not hasattr(field, 'attrib'):
            return None
        if field.attrib.get('name',None) != 'Bitmask':
            return None
        a = str(field).split(',')
        ret = {}
        for v in a:
            a2 = v.split(':')
            if len(a2) == 2:
                ret[a2[0]] = a2[1]
        return ret

    def param_info(self, param, value):
        '''return info string for a param value'''
        htree = self.param_help_tree()
        if htree is None:
            return
        param = param.upper()
        if not param in htree:
            return None
        help = htree[param]
        try:
            bitmask = self.get_bitmask_from_help(help)
            if bitmask is not None:
                v = []
                for k in bitmask.keys():
                    if int(value) & (1<<int(k)):
                        v.append(bitmask[k])
                return '|'.join(v)
        except Exception as e:
            print(e)
            pass
        try:
            values = self.get_Values_from_help(help)
            for v in values:
                if int(v.get('code')) == int(value):
                    return v
        except Exception as e:
            pass
        return None

    def param_help(self, args):
        '''show help on a parameter'''
        if len(args) == 0:
            print("Usage: param help PARAMETER_NAME")
            return

        htree = self.param_help_tree(True)
        if htree is None:
            return

        for h in args:
            h = h.upper()
            if h in htree:
                help = htree[h]
                print("%s: %s\n" % (h, help.get('humanName')))
                print(help.get('documentation'))
                try:
                    print("\n")
                    for f in help.field:
                        print("%s : %s" % (f.get('name'), str(f)))
                except Exception as e:
                    pass
                try:
                    values = self.get_Values_from_help(help)
                    if len(values):
                        print("\nValues: ")
                        for v in values:
                            print("\t%s : %s" % (v.get('code'), str(v)))
                except Exception as e:
                    print("Caught exception %s" % repr(e))
                    pass
            else:
                print("Parameter '%s' not found in documentation" % h)
            
    def param_check(self, params, args):
        '''Check through parameters for obvious misconfigurations'''
        problems_found = False
        htree = self.param_help_tree(True)
        if htree is None:
            return
        for param in params.keys():
            if param.startswith("SIM_"):
                # no documentation for these ATM
                continue
            value = params[param]
#            print("%s: %s" % (param, str(value)))
            try:
                help = htree[param]
            except KeyError:
                print("%s: not found in documentation" % (param,))
                problems_found = True
                continue

            # we'll ignore the Values field if there's a bitmask field
            # involved as they're usually just examples.
            has_bitmask = False
            for f in getattr(help, "field", []):
                if f.get('name') == "Bitmask":
                    has_bitmask = True
                    break
            if not has_bitmask:
                values = self.get_Values_from_help(help)
                if len(values) == 0:
                    # no prescribed values list
                    continue
                value_values = [float(x.get("code")) for x in values]
                if value not in value_values:
                    print("%s: value %f not in Values (%s)" %
                          (param, value, str(value_values)))
                    problems_found = True

        if problems_found:
            print("Remember to `param download` before trusting the checking!  Also, remember that parameter documentation is for *master*!")




"""
  MAVProxy console

  uses lib/console.py for display
"""

import os, sys, math, time, re

from MAVProxy.modules.lib import wxconsole
from MAVProxy.modules.lib import textconsole
from MAVProxy.modules.mavproxy_map import mp_elevation
from pymavlink import mavutil
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import wxsettings
from MAVProxy.modules.lib.mp_menu import *

class DisplayItem:
    def __init__(self, fmt, expression, row):
        self.expression = expression.strip('"\'')
        self.format = fmt.strip('"\'')
        re_caps = re.compile('[A-Z_][A-Z0-9_]+')
        self.msg_types = set(re.findall(re_caps, expression))
        print(self.expression, self.msg_types)
        self.row = row

class ConsoleModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(ConsoleModule, self).__init__(mpstate, "console", "GUI console", public=True, multi_vehicle=True)
        self.in_air = False
        self.start_time = 0.0
        self.total_time = 0.0
        self.speed = 0
        self.max_link_num = 0
        self.last_sys_status_health = 0
        self.last_sys_status_errors_announce = 0
        self.user_added = {}
        self.safety_on = False
        self.unload_check_interval = 5 # seconds
        self.last_unload_check_time = time.time()
        self.add_command('console', self.cmd_console, "console module", ['add','list','remove'])
        mpstate.console = wxconsole.MessageConsole(title='Console')

        # setup some default status information
        mpstate.console.set_status('Mode', 'UNKNOWN', row=0, fg='blue')
        mpstate.console.set_status('SysID', '', row=0, fg='blue')
        mpstate.console.set_status('ARM', 'ARM', fg='grey', row=0)
        mpstate.console.set_status('GPS', 'GPS: --', fg='red', row=0)
        mpstate.console.set_status('GPS2', '', fg='red', row=0)
        mpstate.console.set_status('Vcc', 'Vcc: --', fg='red', row=0)
        mpstate.console.set_status('Radio', 'Radio: --', row=0)
        mpstate.console.set_status('INS', 'INS', fg='grey', row=0)
        mpstate.console.set_status('MAG', 'MAG', fg='grey', row=0)
        mpstate.console.set_status('AS', 'AS', fg='grey', row=0)
        mpstate.console.set_status('RNG', 'RNG', fg='grey', row=0)
        mpstate.console.set_status('AHRS', 'AHRS', fg='grey', row=0)
        mpstate.console.set_status('EKF', 'EKF', fg='grey', row=0)
        mpstate.console.set_status('LOG', 'LOG', fg='grey', row=0)
        mpstate.console.set_status('Heading', 'Hdg ---/---', row=2)
        mpstate.console.set_status('Alt', 'Alt ---', row=2)
        mpstate.console.set_status('AGL', 'AGL ---/---', row=2)
        mpstate.console.set_status('AirSpeed', 'AirSpeed --', row=2)
        mpstate.console.set_status('GPSSpeed', 'GPSSpeed --', row=2)
        mpstate.console.set_status('Thr', 'Thr ---', row=2)
        mpstate.console.set_status('Roll', 'Roll ---', row=2)
        mpstate.console.set_status('Pitch', 'Pitch ---', row=2)
        mpstate.console.set_status('Wind', 'Wind ---/---', row=2)
        mpstate.console.set_status('WP', 'WP --', row=3)
        mpstate.console.set_status('WPDist', 'Distance ---', row=3)
        mpstate.console.set_status('WPBearing', 'Bearing ---', row=3)
        mpstate.console.set_status('AltError', 'AltError --', row=3)
        mpstate.console.set_status('AspdError', 'AspdError --', row=3)
        mpstate.console.set_status('FlightTime', 'FlightTime --', row=3)
        mpstate.console.set_status('ETR', 'ETR --', row=3)
        mpstate.console.set_status('Params', 'Param ---/---', row=3)
        mpstate.console.set_status('Mission', 'Mission --/--', row=3)

        mpstate.console.ElevationMap = mp_elevation.ElevationModel()

        self.vehicle_list = []
        self.vehicle_menu = None
        self.vehicle_name_by_sysid = {}
        self.component_name = {}
        self.last_param_sysid_timestamp = None

        # create the main menu
        if mp_util.has_wxpython:
            self.menu = MPMenuTop([])
            self.add_menu(MPMenuSubMenu('MAVProxy',
                                        items=[MPMenuItem('Settings', 'Settings', 'menuSettings'),
                                               MPMenuItem('Show Map', 'Load Map', '# module load map'),
                                               MPMenuItem('Show HUD', 'Load HUD', '# module load horizon')]))
            self.vehicle_menu = MPMenuSubMenu('Vehicle', items=[])
            self.add_menu(self.vehicle_menu)

    def cmd_console(self, args):
        usage = 'usage: console <add|list|remove>'
        if len(args) < 1:
            print(usage)
            return
        cmd = args[0]
        if cmd == 'add':
            if len(args) < 4:
                print("usage: console add ID FORMAT EXPRESSION <row>")
                return
            if len(args) > 4:
                row = int(args[4])
            else:
                row = 4
            self.user_added[args[1]] = DisplayItem(args[2], args[3], row)
        elif cmd == 'list':
            for k in sorted(self.user_added.keys()):
                d = self.user_added[k]
                print("%s : FMT=%s EXPR=%s ROW=%u" % (k, d.format, d.expression, d.row))
        elif cmd == 'remove':
            if len(args) < 2:
                print("usage: console remove ID")
                return
            id = args[1]
            if id in self.user_added:
                self.user_added.pop(id)
        else:
            print(usage)

    def add_menu(self, menu):
        '''add a new menu'''
        self.menu.add(menu)
        self.mpstate.console.set_menu(self.menu, self.menu_callback)

    def remove_menu(self, menu):
        '''add a new menu'''
        self.menu.remove(menu)
        self.mpstate.console.set_menu(self.menu, self.menu_callback)

    def unload(self):
        '''unload module'''
        self.mpstate.console.close()
        self.mpstate.console = textconsole.SimpleConsole()

    def menu_callback(self, m):
        '''called on menu selection'''
        if m.returnkey.startswith('# '):
            cmd = m.returnkey[2:]
            if m.handler is not None:
                if m.handler_result is None:
                    return
                cmd += m.handler_result
            self.mpstate.functions.process_stdin(cmd)
        if m.returnkey == 'menuSettings':
            wxsettings.WXSettings(self.settings)


    def estimated_time_remaining(self, lat, lon, wpnum, speed):
        '''estimate time remaining in mission in seconds'''
        idx = wpnum
        if wpnum >= self.module('wp').wploader.count():
            return 0
        distance = 0
        done = set()
        while idx < self.module('wp').wploader.count():
            if idx in done:
                break
            done.add(idx)
            w = self.module('wp').wploader.wp(idx)
            if w.command == mavutil.mavlink.MAV_CMD_DO_JUMP:
                idx = int(w.param1)
                continue
            idx += 1
            if (w.x != 0 or w.y != 0) and w.command in [mavutil.mavlink.MAV_CMD_NAV_WAYPOINT,
                                                        mavutil.mavlink.MAV_CMD_NAV_LOITER_UNLIM,
                                                        mavutil.mavlink.MAV_CMD_NAV_LOITER_TURNS,
                                                        mavutil.mavlink.MAV_CMD_NAV_LOITER_TIME,
                                                        mavutil.mavlink.MAV_CMD_NAV_LAND,
                                                        mavutil.mavlink.MAV_CMD_NAV_TAKEOFF]:
                distance += mp_util.gps_distance(lat, lon, w.x, w.y)
                lat = w.x
                lon = w.y
                if w.command == mavutil.mavlink.MAV_CMD_NAV_LAND:
                    break
        return distance / speed

    def vehicle_type_string(self, hb):
        '''return vehicle type string from a heartbeat'''
        if hb.type == mavutil.mavlink.MAV_TYPE_FIXED_WING:
            return 'Plane'
        if hb.type == mavutil.mavlink.MAV_TYPE_GROUND_ROVER:
            return 'Rover'
        if hb.type == mavutil.mavlink.MAV_TYPE_SURFACE_BOAT:
            return 'Boat'
        if hb.type == mavutil.mavlink.MAV_TYPE_SUBMARINE:
            return 'Sub'
        if hb.type in [mavutil.mavlink.MAV_TYPE_QUADROTOR,
                           mavutil.mavlink.MAV_TYPE_COAXIAL,
                           mavutil.mavlink.MAV_TYPE_HEXAROTOR,
                           mavutil.mavlink.MAV_TYPE_OCTOROTOR,
                           mavutil.mavlink.MAV_TYPE_TRICOPTER,
                           mavutil.mavlink.MAV_TYPE_DODECAROTOR]:
            return "Copter"
        if hb.type == mavutil.mavlink.MAV_TYPE_HELICOPTER:
            return "Heli"
        if hb.type == mavutil.mavlink.MAV_TYPE_ANTENNA_TRACKER:
            return "Tracker"
        if hb.type == mavutil.mavlink.MAV_TYPE_AIRSHIP:
            return "Blimp"
        elif hb.type == mavutil.mavlink.MAV_TYPE_ADSB:
            return "ADSB"
        return "UNKNOWN(%u)" % hb.type

    def component_type_string(self, hb):
        # note that we rely on vehicle_type_string for basic vehicle types
        if hb.type == mavutil.mavlink.MAV_TYPE_GCS:
            return "GCS"
        elif hb.type == mavutil.mavlink.MAV_TYPE_GIMBAL:
            return "Gimbal"
        elif hb.type == mavutil.mavlink.MAV_TYPE_ONBOARD_CONTROLLER:
            return "CC"
        elif hb.type == mavutil.mavlink.MAV_TYPE_ADSB:
            return "ADSB"
        elif hb.type == mavutil.mavlink.MAV_TYPE_GENERIC:
            return "Generic"
        return self.vehicle_type_string(hb)

    def update_vehicle_menu(self):
        '''update menu for new vehicles'''
        self.vehicle_menu.items = []
        for s in sorted(self.vehicle_list):
            clist = self.module('param').get_component_id_list(s)
            if len(clist) == 1:
                name = 'SysID %u: %s' % (s, self.vehicle_name_by_sysid[s])
                self.vehicle_menu.items.append(MPMenuItem(name, name, '# vehicle %u' % s))
            else:
                for c in sorted(clist):
                    try:
                        name = 'SysID %u[%u]: %s' % (s, c, self.component_name[s][c])
                    except KeyError as e:
                        name = 'SysID %u[%u]: ?' % (s,c)
                    self.vehicle_menu.items.append(MPMenuItem(name, name, '# vehicle %u:%u' % (s,c)))
        self.mpstate.console.set_menu(self.menu, self.menu_callback)
    
    def add_new_vehicle(self, hb):
        '''add a new vehicle'''
        if hb.type == mavutil.mavlink.MAV_TYPE_GCS:
            return
        sysid = hb.get_srcSystem()
        self.vehicle_list.append(sysid)
        self.vehicle_name_by_sysid[sysid] = self.vehicle_type_string(hb)
        self.update_vehicle_menu()

    def check_critical_error(self, msg):
        '''check for any error bits being set in SYS_STATUS'''
        errors = msg.errors_count1 | (msg.errors_count2<<16)
        if errors == 0:
            return
        now = time.time()
        if now - self.last_sys_status_errors_announce > self.mpstate.settings.sys_status_error_warn_interval:
            self.last_sys_status_errors_announce = now
            self.say("Critical failure 0x%x sysid=%u" % (errors, msg.get_srcSystem()))

        
    def mavlink_packet(self, msg):
        '''handle an incoming mavlink packet'''
        if not isinstance(self.console, wxconsole.MessageConsole):
            return
        if not self.console.is_alive():
            self.mpstate.console = textconsole.SimpleConsole()
            return
        type = msg.get_type()
        sysid = msg.get_srcSystem()
        compid = msg.get_srcComponent()

        if type == 'HEARTBEAT' or type == 'HIGH_LATENCY2':
            if not sysid in self.vehicle_list:
                self.add_new_vehicle(msg)
            if sysid not in self.component_name:
                self.component_name[sysid] = {}
            if compid not in self.component_name[sysid]:
                self.component_name[sysid][compid] = self.component_type_string(msg)
                self.update_vehicle_menu()

        if self.last_param_sysid_timestamp != self.module('param').new_sysid_timestamp:
            '''a new component ID has appeared for parameters'''
            self.last_param_sysid_timestamp = self.module('param').new_sysid_timestamp
            self.update_vehicle_menu()

        if type in ['RADIO', 'RADIO_STATUS']:
            # handle RADIO msgs from all vehicles
            if msg.rssi < msg.noise+10 or msg.remrssi < msg.remnoise+10:
                fg = 'red'
            else:
                fg = 'black'
            self.console.set_status('Radio', 'Radio %u/%u %u/%u' % (msg.rssi, msg.noise, msg.remrssi, msg.remnoise), fg=fg)

        if type == 'SYS_STATUS':
            self.check_critical_error(msg)

        if not self.is_primary_vehicle(msg):
            # don't process msgs from other than primary vehicle, other than
            # updating vehicle list
            return

        master = self.master
        # add some status fields
        if type in [ 'GPS_RAW_INT', 'GPS2_RAW' ]:
            if type == 'GPS_RAW_INT':
                field = 'GPS'
                prefix = 'GPS:'
            else:
                field = 'GPS2'
                prefix = 'GPS2'
            nsats = msg.satellites_visible
            fix_type = msg.fix_type
            if fix_type >= 3:
                self.console.set_status(field, '%s OK%s (%u)' % (prefix, fix_type, nsats), fg='green')
            else:
                self.console.set_status(field, '%s %u (%u)' % (prefix, fix_type, nsats), fg='red')
            if type == 'GPS_RAW_INT':
                vfr_hud_heading = master.field('VFR_HUD', 'heading', None)
                gps_heading = int(msg.cog * 0.01)
                if vfr_hud_heading is None:
                    vfr_hud_heading = '---'
                else:
                    vfr_hud_heading = '%3u' % vfr_hud_heading
                self.console.set_status('Heading', 'Hdg %s/%3u' %
                                        (vfr_hud_heading, gps_heading))
        elif type == 'VFR_HUD':
            if master.mavlink10():
                alt = master.field('GPS_RAW_INT', 'alt', 0) / 1.0e3
            else:
                alt = master.field('GPS_RAW', 'alt', 0)
            home = self.module('wp').get_home()
            if home is not None:
                home_lat = home.x
                home_lng = home.y
            else:
                home_lat = None
                home_lng = None
            lat = master.field('GLOBAL_POSITION_INT', 'lat', 0) * 1.0e-7
            lng = master.field('GLOBAL_POSITION_INT', 'lon', 0) * 1.0e-7
            rel_alt = master.field('GLOBAL_POSITION_INT', 'relative_alt', 0) * 1.0e-3
            agl_alt = None
            if self.settings.basealt != 0:
                agl_alt = self.console.ElevationMap.GetElevation(lat, lng)
                if agl_alt is not None:
                    agl_alt = self.settings.basealt - agl_alt
            else:
                try:
                    agl_alt_home = self.console.ElevationMap.GetElevation(home_lat, home_lng)
                except Exception as ex:
                    print(ex)
                    agl_alt_home = None
                if agl_alt_home is not None:
                    agl_alt = self.console.ElevationMap.GetElevation(lat, lng)
                if agl_alt is not None:
                    agl_alt = agl_alt_home - agl_alt
            if agl_alt is not None:
                agl_alt += rel_alt
                vehicle_agl = master.field('TERRAIN_REPORT', 'current_height', None)
                if vehicle_agl is None:
                    vehicle_agl = '---'
                else:
                    vehicle_agl = self.height_string(vehicle_agl)
                self.console.set_status('AGL', 'AGL %s/%s' % (self.height_string(agl_alt), vehicle_agl))
            self.console.set_status('Alt', 'Alt %s' % self.height_string(rel_alt))
            self.console.set_status('AirSpeed', 'AirSpeed %s' % self.speed_string(msg.airspeed))
            self.console.set_status('GPSSpeed', 'GPSSpeed %s' % self.speed_string(msg.groundspeed))
            self.console.set_status('Thr', 'Thr %u' % msg.throttle)
            t = time.localtime(msg._timestamp)
            flying = False
            if self.mpstate.vehicle_type == 'copter':
                flying = self.master.motors_armed()
            else:
                flying = msg.groundspeed > 3
            if flying and not self.in_air:
                self.in_air = True
                self.start_time = time.mktime(t)
            elif flying and self.in_air:
                self.total_time = time.mktime(t) - self.start_time
                self.console.set_status('FlightTime', 'FlightTime %u:%02u' % (int(self.total_time)/60, int(self.total_time)%60))
            elif not flying and self.in_air:
                self.in_air = False
                self.total_time = time.mktime(t) - self.start_time
                self.console.set_status('FlightTime', 'FlightTime %u:%02u' % (int(self.total_time)/60, int(self.total_time)%60))
        elif type == 'ATTITUDE':
            self.console.set_status('Roll', 'Roll %u' % math.degrees(msg.roll))
            self.console.set_status('Pitch', 'Pitch %u' % math.degrees(msg.pitch))
        elif type in ['SYS_STATUS']:
            sensors = { 'AS'   : mavutil.mavlink.MAV_SYS_STATUS_SENSOR_DIFFERENTIAL_PRESSURE,
                        'MAG'  : mavutil.mavlink.MAV_SYS_STATUS_SENSOR_3D_MAG,
                        'INS'  : mavutil.mavlink.MAV_SYS_STATUS_SENSOR_3D_ACCEL | mavutil.mavlink.MAV_SYS_STATUS_SENSOR_3D_GYRO,
                        'AHRS' : mavutil.mavlink.MAV_SYS_STATUS_AHRS,
                        'RC'   : mavutil.mavlink.MAV_SYS_STATUS_SENSOR_RC_RECEIVER,
                        'TERR' : mavutil.mavlink.MAV_SYS_STATUS_TERRAIN,
                        'RNG'  : mavutil.mavlink.MAV_SYS_STATUS_SENSOR_LASER_POSITION,
                        'LOG'  : mavutil.mavlink.MAV_SYS_STATUS_LOGGING,
                        'PRX'  : mavutil.mavlink.MAV_SYS_STATUS_SENSOR_PROXIMITY,
                        'PRE'  : mavutil.mavlink.MAV_SYS_STATUS_PREARM_CHECK,
            }
            hide_if_not_present = set(['PRE', 'PRX'])
            for s in sensors.keys():
                bits = sensors[s]
                present = ((msg.onboard_control_sensors_present & bits) == bits)
                enabled = ((msg.onboard_control_sensors_enabled & bits) == bits)
                healthy = ((msg.onboard_control_sensors_health & bits) == bits)
                if not present and s in hide_if_not_present:
                    continue
                if not present:
                    fg = 'black'
                elif not enabled:
                    fg = 'grey'
                elif not healthy:
                    fg = 'red'
                else:
                    fg = 'green'
                # for terrain show yellow if still loading
                if s == 'TERR' and fg == 'green' and master.field('TERRAIN_REPORT', 'pending', 0) != 0:
                    fg = 'yellow'
                self.console.set_status(s, s, fg=fg)
            announce_unhealthy = {
                'RC': 'RC',
                'PRE': 'pre-arm',
            }
            for s in announce_unhealthy.keys():
                bits = sensors[s]
                enabled = ((msg.onboard_control_sensors_enabled & bits) == bits)
                healthy = ((msg.onboard_control_sensors_health & bits) == bits)
                was_healthy = ((self.last_sys_status_health & bits) == bits)
                if enabled and not healthy and was_healthy:
                    self.say("%s fail" % announce_unhealthy[s])
            announce_healthy = {
                'PRE': 'pre-arm',
            }
            for s in announce_healthy.keys():
                bits = sensors[s]
                enabled = ((msg.onboard_control_sensors_enabled & bits) == bits)
                healthy = ((msg.onboard_control_sensors_health & bits) == bits)
                was_healthy = ((self.last_sys_status_health & bits) == bits)
                if enabled and healthy and not was_healthy:
                    self.say("%s good" % announce_healthy[s])
            self.last_sys_status_health = msg.onboard_control_sensors_health

            if ((msg.onboard_control_sensors_enabled & mavutil.mavlink.MAV_SYS_STATUS_SENSOR_MOTOR_OUTPUTS) == 0):
                self.safety_on = True
            else:
                self.safety_on = False                

        elif type == 'WIND':
            self.console.set_status('Wind', 'Wind %u/%s' % (msg.direction, self.speed_string(msg.speed)))

        elif type == 'EKF_STATUS_REPORT':
            highest = 0.0
            vars = ['velocity_variance',
                    'pos_horiz_variance',
                    'pos_vert_variance',
                    'compass_variance',
                    'terrain_alt_variance']
            for var in vars:
                v = getattr(msg, var, 0)
                highest = max(v, highest)
            if highest >= 1.0:
                fg = 'red'
            elif highest >= 0.5:
                fg = 'orange'
            else:
                fg = 'green'
            self.console.set_status('EKF', 'EKF', fg=fg)

        elif type == 'HWSTATUS':
            if msg.Vcc >= 4600 and msg.Vcc <= 5300:
                fg = 'green'
            else:
                fg = 'red'
            self.console.set_status('Vcc', 'Vcc %.2f' % (msg.Vcc * 0.001), fg=fg)
        elif type == 'POWER_STATUS':
            if msg.flags & mavutil.mavlink.MAV_POWER_STATUS_CHANGED:
                fg = 'red'
            else:
                fg = 'green'
            status = 'PWR:'
            if msg.flags & mavutil.mavlink.MAV_POWER_STATUS_USB_CONNECTED:
                status += 'U'
            if msg.flags & mavutil.mavlink.MAV_POWER_STATUS_BRICK_VALID:
                status += 'B'
            if msg.flags & mavutil.mavlink.MAV_POWER_STATUS_SERVO_VALID:
                status += 'S'
            if msg.flags & mavutil.mavlink.MAV_POWER_STATUS_PERIPH_OVERCURRENT:
                status += 'O1'
            if msg.flags & mavutil.mavlink.MAV_POWER_STATUS_PERIPH_HIPOWER_OVERCURRENT:
                status += 'O2'
            self.console.set_status('PWR', status, fg=fg)
            self.console.set_status('Srv', 'Srv %.2f' % (msg.Vservo*0.001), fg='green')
        elif type in ['HEARTBEAT', 'HIGH_LATENCY2']:
            fmode = master.flightmode
            if self.settings.vehicle_name:
                fmode = self.settings.vehicle_name + ':' + fmode
            self.console.set_status('Mode', '%s' % fmode, fg='blue')
            if len(self.vehicle_list) > 1:
                self.console.set_status('SysID', 'Sys:%u' % sysid, fg='blue')
            if self.master.motors_armed():
                arm_colour = 'green'
            else:
                arm_colour = 'red'
            armstring = 'ARM'
            # add safety switch state
            if self.safety_on:
                armstring += '(SAFE)'
            self.console.set_status('ARM', armstring, fg=arm_colour)
            if self.max_link_num != len(self.mpstate.mav_master):
                for i in range(self.max_link_num):
                    self.console.set_status('Link%u'%(i+1), '', row=1)
                self.max_link_num = len(self.mpstate.mav_master)
            for m in self.mpstate.mav_master:
                if self.mpstate.settings.checkdelay:
                    highest_msec_key = (sysid, compid)
                    linkdelay = (self.mpstate.status.highest_msec.get(highest_msec_key, 0) - m.highest_msec.get(highest_msec_key,0))*1.0e-3
                else:
                    linkdelay = 0
                linkline = "Link %s " % (self.link_label(m))
                fg = 'dark green'
                if m.linkerror:
                    linkline += "down"
                    fg = 'red'
                else:
                    packets_rcvd_percentage = 100
                    if (m.mav_count+m.mav_loss) != 0: #avoid divide-by-zero
                        packets_rcvd_percentage = (100.0 * m.mav_count) / (m.mav_count + m.mav_loss)

                    linkbits = ["%u pkts" % m.mav_count,
                                "%u lost" % m.mav_loss,
                                "%.2fs delay" % linkdelay,
                    ]
                    try:
                        if m.mav.signing.sig_count:
                            # other end is sending us signed packets
                            if not m.mav.signing.secret_key:
                                # we've received signed packets but
                                # can't verify them
                                fg = 'orange'
                                linkbits.append("!KEY")
                            elif not m.mav.signing.sign_outgoing:
                                # we've received signed packets but aren't
                                # signing outselves; this can lead to hairloss
                                fg = 'orange'
                                linkbits.append("!SIGNING")
                            if m.mav.signing.badsig_count:
                                fg = 'orange'
                                linkbits.append("%u badsigs" % m.mav.signing.badsig_count)
                    except AttributeError as e:
                        # mav.signing.sig_count probably doesn't exist
                        pass

                    linkline += "OK {rcv_pct:.1f}% ({bits})".format(
                        rcv_pct=packets_rcvd_percentage,
                        bits=", ".join(linkbits))

                    if linkdelay > 1 and fg == 'dark green':
                        fg = 'orange'

                self.console.set_status('Link%u'%m.linknum, linkline, row=1, fg=fg)
        elif type in ['WAYPOINT_CURRENT', 'MISSION_CURRENT']:
            wpmax = self.module('wp').wploader.count()
            if wpmax > 0:
                wpmax = "/%u" % wpmax
            else:
                wpmax = ""
            self.console.set_status('WP', 'WP %u%s' % (msg.seq, wpmax))
            lat = master.field('GLOBAL_POSITION_INT', 'lat', 0) * 1.0e-7
            lng = master.field('GLOBAL_POSITION_INT', 'lon', 0) * 1.0e-7
            if lat != 0 and lng != 0:
                airspeed = master.field('VFR_HUD', 'airspeed', 30)
                if abs(airspeed - self.speed) > 5:
                    self.speed = airspeed
                else:
                    self.speed = 0.98*self.speed + 0.02*airspeed
                self.speed = max(1, self.speed)
                time_remaining = int(self.estimated_time_remaining(lat, lng, msg.seq, self.speed))
                self.console.set_status('ETR', 'ETR %u:%02u' % (time_remaining/60, time_remaining%60))

        elif type == 'NAV_CONTROLLER_OUTPUT':
            self.console.set_status('WPDist', 'Distance %s' % self.dist_string(msg.wp_dist))
            self.console.set_status('WPBearing', 'Bearing %u' % msg.target_bearing)
            if msg.alt_error > 0:
                alt_error_sign = "(L)"
            else:
                alt_error_sign = "(H)"
            if msg.aspd_error > 0:
                aspd_error_sign = "(L)"
            else:
                aspd_error_sign = "(H)"
            if math.isnan(msg.alt_error):
                alt_error = "NaN"
            else:
                alt_error = "%s%s" % (self.height_string(msg.alt_error), alt_error_sign)
            self.console.set_status('AltError', 'AltError %s' % alt_error)
            self.console.set_status('AspdError', 'AspdError %s%s' % (self.speed_string(msg.aspd_error*0.01), aspd_error_sign))

        elif type == 'PARAM_VALUE':
            rec, tot = self.module('param').param_status()
            self.console.set_status('Params', 'Param %u/%u' % (rec,tot))
            
        if type == 'HIGH_LATENCY2':
            self.console.set_status('WPDist', 'Distance %s' % self.dist_string(msg.target_distance * 10))
            # The -180 here for for consistency with NAV_CONTROLLER_OUTPUT (-180->180), whereas HIGH_LATENCY2 is (0->360)
            self.console.set_status('WPBearing', 'Bearing %u' % ((msg.target_heading * 2) - 180))
            alt_error = "%s%s" % (self.height_string(msg.target_altitude - msg.altitude),
                                  "(L)" if (msg.target_altitude - msg.altitude) > 0 else "(L)")
            self.console.set_status('AltError', 'AltError %s' % alt_error)
            self.console.set_status('AspdError', 'AspdError %s%s' % (self.speed_string((msg.airspeed_sp - msg.airspeed)/5),
                                                                    "(L)" if (msg.airspeed_sp - msg.airspeed) > 0 else "(L)"))
            # The -180 here for for consistency with WIND (-180->180), whereas HIGH_LATENCY2 is (0->360)
            self.console.set_status('Wind', 'Wind %u/%s' % ((msg.wind_heading * 2) - 180, self.speed_string(msg.windspeed / 5)))
            self.console.set_status('Alt', 'Alt %s' % self.height_string(msg.altitude - self.console.ElevationMap.GetElevation(msg.latitude / 1E7, msg.longitude / 1E7)))
            self.console.set_status('AirSpeed', 'AirSpeed %s' % self.speed_string(msg.airspeed / 5))
            self.console.set_status('GPSSpeed', 'GPSSpeed %s' % self.speed_string(msg.groundspeed / 5))
            self.console.set_status('Thr', 'Thr %u' % msg.throttle)
            self.console.set_status('Heading', 'Hdg %s/---' % (msg.heading * 2))
            self.console.set_status('WP', 'WP %u/--' % (msg.wp_num))
            
            #re-map sensors
            sensors = { 'AS'   : mavutil.mavlink.HL_FAILURE_FLAG_DIFFERENTIAL_PRESSURE,
                        'MAG'  : mavutil.mavlink.HL_FAILURE_FLAG_3D_MAG,
                        'INS'  : mavutil.mavlink.HL_FAILURE_FLAG_3D_ACCEL | mavutil.mavlink.HL_FAILURE_FLAG_3D_GYRO,
                        'AHRS' : mavutil.mavlink.HL_FAILURE_FLAG_ESTIMATOR,
                        'RC'   : mavutil.mavlink.HL_FAILURE_FLAG_RC_RECEIVER,
                        'TERR' : mavutil.mavlink.HL_FAILURE_FLAG_TERRAIN
            }
            for s in sensors.keys():
                bits = sensors[s]
                failed = ((msg.failure_flags & bits) == bits)
                if failed:
                    fg = 'red'
                else:
                    fg = 'green'
                self.console.set_status(s, s, fg=fg)
                
            # do the remaining non-standard system mappings
            fence_failed = ((msg.failure_flags & mavutil.mavlink.HL_FAILURE_FLAG_GEOFENCE) == mavutil.mavlink.HL_FAILURE_FLAG_GEOFENCE)
            if fence_failed:
                fg = 'red'
            else:
                fg = 'green'
            self.console.set_status('Fence', 'FEN', fg=fg)
            gps_failed = ((msg.failure_flags & mavutil.mavlink.HL_FAILURE_FLAG_GPS) == mavutil.mavlink.HL_FAILURE_FLAG_GPS)
            if gps_failed:
                self.console.set_status('GPS', 'GPS FAILED', fg='red')
            else:
                self.console.set_status('GPS', 'GPS OK', fg='green')
            batt_failed = ((msg.failure_flags & mavutil.mavlink.HL_FAILURE_FLAG_GPS) == mavutil.mavlink.HL_FAILURE_FLAG_BATTERY)
            if batt_failed:
                self.console.set_status('PWR', 'PWR FAILED', fg='red')
            else:
                self.console.set_status('PWR', 'PWR OK', fg='green')            
                                                                        
        for id in self.user_added.keys():
            if type in self.user_added[id].msg_types:
                d = self.user_added[id]
                val = mavutil.evaluate_expression(d.expression, self.master.messages)
                try:
                    self.console.set_status(id, d.format % val, row = d.row)
                except Exception as ex:
                    print(ex)
                    pass

    def idle_task(self):
        now = time.time()
        if self.last_unload_check_time + self.unload_check_interval < now:
            self.last_unload_check_time = now
            if not self.console.is_alive():
                self.needs_unloading = True

def init(mpstate):
    '''initialise module'''
    return ConsoleModule(mpstate)



#!/usr/bin/env python
'''
movinghome module
André Kjellstrup, Norce

This module can update the home position of the ArduPilot vehicle to the position of a moving GCS.
requires package python-nmea2

'''

import os
import os.path
import sys
from pymavlink import mavutil
import errno
import time
import math
import serial
try:
    import pynmea2
except ImportError as e:
    print("\n!!! missing package !!! do: 'sudo apt install python-nmea2' -movinghome will not work without it")
from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import mp_settings


class movinghome(mp_module.MPModule):
    def __init__(self, mpstate):
        #Initialise module
        super(movinghome, self).__init__(mpstate, "movinghome", "")
        #latest GCS coordinates
        self.lat = 0
        self.lon = 0
        self.alt = 0
        #last/set home coordinates
        self.lath = 0
        self.lonh = 0  
        self.alth = 0
        #misc settings
        self.device = "/dev/ttyUSB0"
        self.baud = 4800
        self.updating=False
        self.radius = 15 #min travelled distance (m) before update
        self.check_interval = 3 # seconds
        self.last_check = time.time()
        self.fresh = True #fresh start/first movement
        self.dist = 0
        self.last_decode_error_print = 0

        print("\nDefault NMEA source is: %s at %s baud, change if needed before turning on." % (self.device , self.baud))
        self.add_command('movinghome', self.cmd_movinghome, "movinghome module")

            

    def cmd_movinghome(self, args):
        '''control behaviour of the module'''
        if len(args) == 0:
            print("Usage: movinghome <status|on|off|radius|device|baud>\n On/Off: enable position update.\n Radius defines the threshold in meters (2D) from last position, when exceeded, home position is updated.\n device and baud is the serial port setup for NMEA device.")
        elif args[0] == "status":
            self.status()
        elif args[0] == "on":
            self.movinghome_on()
        elif args[0] == "off":
            self.movinghome_off()
        elif args[0] == "radius":
            if len(args) < 2:
                print("Usage: moving base minimum travel radius <RADIUS>")
                return
            self.radius=float(args[1])
        elif args[0] == "device":
            if len(args) < 2:
                print("Usage: device name; device /dev/ttyUSB2")
                return
            self.device=args[1]
        elif args[0] == "baud":
            if len(args) < 2:
                print("Usage: baud rate; baud 9600")
                return
            self.baud=args[1]
        else:
            print(self.usage)

    def status(self):
        #Returns information about module'''
        if self.updating == True:
            print("Last known GCS position lat %(lat)f lon=%(lon)f  max radius=%(max).1fm" %
                   {"lat": self.lath,
                    "lon": self.lonh,
                    "max": self.radius,
                   })
        else:
            print("Not updating home")
        print("Radius is %sm \nInterval is %ss \nDevice is %s at %s baud\n" % (self.radius, self.check_interval, self.device, self.baud ))


    def movinghome_on(self):
        #self.ser = serial.Serial('/dev/ttyUSB0',4800)
        self.ser = serial.Serial(self.device,self.baud)
        self.updating=True
        self.lath = 0 # ensure push of current home.
        print("Home position will be updated if GCS moves > %(max).1fm" %
               {"max": self.radius,
               })

    def movinghome_off(self):
        self.updating = False
        print("Home position will not be updated.")

    def idle_task(self):
        #Called frequently by mavproxy
        if self.updating == True:
            data = self.ser.readline()
            try:
                data = data.decode("ascii")
            except UnicodeDecodeError as e:
                # this is probably a baudrate issue
                now = time.time()
                if now - self.last_decode_error_print > 10:
                    print("movinghome: decode error; baudrate issue?")
                    self.last_decode_error_print = now
                return
            if (data.startswith("$GPGGA")):
                msg = pynmea2.parse(data)
                if int(msg.num_sats) > 5:
                    #convert LAT
                    DD = int(float(msg.lat)/100)
                    MM = float(msg.lat) - DD * 100
                    self.lat = DD + MM/60
                    if msg.lat_dir == "S":
                        self.lat = -self.lat;
                    #convert LON
                    DD = int(float(msg.lon)/100)
                    MM = float(msg.lon) - DD * 100
                    self.lon = DD + MM/60
                    if msg.lon_dir == "W":
                        self.lon = -self.lon;

                now = time.time()
                if now-self.last_check > self.check_interval:
                    self.last_check = now
                    #check if we moved enough
                    self.dist = self.haversine(self.lon, self.lat, self.alt, self.lonh, self.lath, self.alth)
                    if self.dist > self.radius:
                        if self.fresh == True:
                            self.say("GCS position set as home")    
                            self.fresh = False
                        else:
                            message = "GCS moved "
                            message2 = message + "%.0f" % self.dist + "meters"
                            self.say("%s: %s" % (self.name,message2))
                            message2_enc = message2.encode(bytes)
                            self.master.mav.statustext_send(mavutil.mavlink.MAV_SEVERITY_NOTICE, message2)
                        self.console.writeln("home position updated")

                        self.master.mav.command_int_send(
                        self.settings.target_system, self.settings.target_component,
                        mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT_INT,
                        mavutil.mavlink.MAV_CMD_DO_SET_HOME,
                        1, # (1, set current location as home)
                        0, # move on
                        0, # param1
                        0, # param2
                        0, # param3
                        0, # param4
                        int(self.lat*1e7), # param5
                        int(self.lon*1e7), # param6
                        0) # param7

                        self.lath = self.lat
                        self.lonh = self.lon
                        #print data
                        self.console.writeln("%s: %s %s GNSS Quality %s Sats %s "% (self.name,self.lat,self.lon,msg.gps_qual,msg.num_sats))


    def haversine(self, lon1, lat1, alt1, lon2, lat2, alt2):
        r_earth = 6371000
        lon1, lat1, lon2, lat2 = map(math.radians, [lon1, lat1, lon2, lat2])
        dlon = lon2 - lon1
        dlat = lat2 - lat1
        a = math.sin(dlat/2)**2 + math.cos(lat1) * math.cos(lat2) * math.sin(dlon/2)**2
        c = 2 * math.asin(math.sqrt(a))
        d = c*r_earth
        return math.sqrt(d**2+(alt1 - alt2)**2)


def init(mpstate):
    return movinghome(mpstate)



# Copyright (C) 2016  Intel Corporation. All rights reserved.
#
# This file is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
'''
Module with helpers for OpenGL rendering.
'''
import math

from ctypes import *
from OpenGL.GL import *
from pymavlink.quaternion import Quaternion
from pymavlink.rotmat import Vector3

def Vector3_to_tuple(v):
    return (v.x, v.y, v.z)

class Material(object):
    def __init__(self,
                 ambient=Vector3(.5, .5, .5),
                 diffuse=Vector3(.5, .5, .5),
                 specular=Vector3(.5, .5, .5),
                 specular_exponent=32.0,
                 alpha=1.0):
        self.ambient = ambient
        self.diffuse = diffuse
        self.specular = specular
        self.specular_exponent = specular_exponent
        self.alpha = alpha

    def set_color(self, color):
        self.ambient = color
        self.diffuse = color
        self.specular = color

class Light(object):
    def __init__(self,
                 position=Vector3(0.0, 0.0, 1.0),
                 ambient=Vector3(1.0, 1.0, 1.0),
                 diffuse=Vector3(1.0, 1.0, 1.0),
                 specular=Vector3(1.0, 1.0, 1.0),
                 att_linear=0.0,
                 att_quad=0.0):
        self.position = position
        self.ambient = ambient
        self.diffuse = diffuse
        self.specular = specular
        self.att_linear = att_linear
        self.att_quad = att_quad

class Transform(object):
    '''Class to represent transform operations. Note that the operations
    provided are isolated from each other, in the sense that the sequence of
    operations is always: rotation, scale and translation. That means that an
    operation add to itself and the final outcome will always be in the
    following order: accumulated scale, accumulated rotation and accumulated
    translation.'''
    def __init__(self):
        self.quaternion = Quaternion((1, 0, 0, 0))
        self.translation = Vector3(0.0, 0.0, 0.0)
        self.scale_factor = 1.0

    def scale(self, scale):
        self.scale_factor *= scale

    def rotation_quaternion(self, vector, angle):
        if not vector.length():
            return None
        c = math.cos(angle / 2.0)
        v = math.sin(angle / 2.0) * vector.normalized()
        q = Quaternion((c, v.x, v.y, v.z))
        q.normalize()
        return q

    def rotate(self, vector, angle):
        q = self.rotation_quaternion(vector, angle)
        if not q:
            return
        self.quaternion = q * self.quaternion
        self.quaternion.normalize()

    def set_rotation(self, vector, angle):
        q = self.rotation_quaternion(vector, angle)
        if not q:
            return
        self.quaternion = q

    def set_euler(self, roll, pitch, yaw):
        self.quaternion = Quaternion((roll, pitch, yaw))
        self.quaternion.normalize()

    def translate(self, d):
        self.translation += d

    def mat4(self):
        s = self.scale_factor
        m = self.quaternion.dcm
        d = self.translation
        return (c_float * 16)(
            s * m.a.x, s * m.b.x, s * m.c.x, 0,
            s * m.a.y, s * m.b.y, s * m.c.y, 0,
            s * m.a.z, s * m.b.z, s * m.c.z, 0,
                  d.x,       d.y,       d.z, 1
        )

    def apply(self, v):
        v = self.quaternion.transform(v) * self.scale_factor
        v += self.translation
        return v

class Camera(object):
    def __init__(self):
        # Base must be orthonormal
        self.base = (
            Vector3(1.0, 0.0, 0.0),
            Vector3(0.0, 1.0, 0.0),
            Vector3(0.0, 0.0, 1.0)
        )
        self.position = Vector3(0.0, 0.0, 0.0)
        self.position_transform = Transform()
        self.base_transform = Transform()

    def view_mat4(self):
        p = self.position_transform.apply(self.position)
        i = self.base_transform.apply(self.base[0])
        j = self.base_transform.apply(self.base[1])
        k = self.base_transform.apply(self.base[2])
        tx, ty, tz = p * i, p * j, p * k

        return (c_float * 16)(
            i.x, j.x, k.x, 0,
            i.y, j.y, k.y, 0,
            i.z, j.z, k.z, 0,
            -tx, -ty, -tz, 1
        )

class Orthographic(object):
    def __init__(self, near=0.0, far=2.0, top=1.0, right=1.0):
        self.near = near
        self.far = far
        self.top = top
        self.right = right

    def proj_mat4(self):
        n = self.near
        f = self.far
        t = self.top
        r = self.right
        return (c_float * 16)(
            1 / r, 0, 0, 0,
            0, 1 / t, 0, 0,
            0, 0, -2 / (f - n), 0,
            0, 0, -(f + n) / (f - n), 1
        )


class Perspective(object):
    def __init__(self, near=0.01, far=100.0, top=0.24, right=0.24):
        self.near = near
        self.far = far
        self.top = top
        self.right = right

    def proj_mat4(self):
        n = self.near
        f = self.far
        t = self.top
        r = self.right
        return (c_float * 16)(
            n / r, 0, 0, 0,
            0, n / t, 0, 0,
            0, 0, -(f + n) / (f - n), -1,
            0, 0, -2 * f * n / (f - n), 0
        )

class Object(object):
    def __init__(self, vertices,
                 normals=[],
                 indices=[],
                 material=Material(),
                 vertices_per_face=3,
                 enable_alpha=False):
        self.local = Transform()
        self.model = Transform()
        self.num_vertices = len(vertices)

        assert(self.num_vertices > 0)

        for i in range(self.num_vertices):
            if not isinstance(vertices[i], Vector3):
                vertices[i] = Vector3(*vertices[i])

        self.midpoint = Vector3(0, 0, 0)
        for v in vertices:
            self.midpoint += v
        self.midpoint /= self.num_vertices

        self.material = material

        if normals:
            for i in range(len(normals)):
                if not isinstance(normals[i], Vector3):
                    normals[i] = Vector3(*normals[i])
        else:
            normals = self.calc_normals(vertices)

        if not indices:
            indices = self.calc_indices(len(vertices), vertices_per_face)

        self.num_indices = len(indices)
        self.centroids = None
        if enable_alpha:
            self.centroids = self.calc_centroids(indices, vertices)

        self.vao = self.create_vao(vertices, normals, indices)

    def create_vao(self, vertices, normals, indices):
        vao = glGenVertexArrays(1)
        glBindVertexArray(vao)

        vbo = glGenBuffers(1)
        glBindBuffer(GL_ARRAY_BUFFER, vbo)

        vertices = [x for p in vertices for x in (p.x, p.y, p.z)]
        normals = [x for p in normals for x in (p.x, p.y, p.z)]

        data = vertices + normals
        data = (c_float * len(data))(*data)

        glBufferData(GL_ARRAY_BUFFER, data, GL_STATIC_DRAW);

        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, c_void_p(0))
        glEnableVertexAttribArray(0)

        glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0,
                              c_void_p(sizeof(c_float) * len(vertices)))
        glEnableVertexAttribArray(1)

        ebo = glGenBuffers(1)
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ebo)
        data = (c_int * len(indices))(*indices)
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, data, GL_STATIC_DRAW);

        glBindVertexArray(0)

        return vao

    def calc_centroids(self, indices, vertices):
        num_faces = len(indices) / 3
        centroids = []
        for i in range(num_faces):
            a = vertices[indices[i * 3]]
            b = vertices[indices[i * 3 + 1]]
            c = vertices[indices[i * 3 + 2]]
            centroids.append((a + b + c) / 3)
        return centroids

    def before_draw(self, program, enable_alpha=False):
        glUseProgram(program.program_id)
        program.use_local_transform(self.local)
        program.use_model_transform(self.model)
        if self.material:
            program.use_material(self.material, enable_alpha=enable_alpha)
        glBindVertexArray(self.vao)

    def after_draw(self, program):
        glBindVertexArray(0)

    def draw(self, program, faces=None, camera=None):
        has_alpha = self.material and self.centroids and camera
        self.before_draw(program, has_alpha)

        if has_alpha:
            if faces is None:
                num_faces = self.num_indices / 3
                faces = [i for i in range(num_faces)]
            # Sort faces based on distance between centroids and camera position
            dists = {}
            for i in faces:
                v = self.model.apply(self.local.apply(self.centroids[i]))
                dists[i] = (camera.position - v).length()
            faces = sorted(faces, None, lambda i: dists[i])
        elif faces:
            faces = sorted(faces)

        if faces is None:
            glDrawElements(GL_TRIANGLES, self.num_indices, GL_UNSIGNED_INT, None)
        elif faces:
            i0 = 0
            l = len(faces)
            i = 1
            while i < l:
                if faces[i] != faces[i - 1] + 1:
                    pointer = c_void_p(sizeof(c_uint) * faces[i0] * 3)
                    glDrawElements(GL_TRIANGLES, (i - i0) * 3, GL_UNSIGNED_INT,
                                   pointer)
                    i0 = i
                i += 1
            pointer = c_void_p(sizeof(c_uint) * faces[i0] * 3)
            glDrawElements(GL_TRIANGLES, (i - i0) * 3, GL_UNSIGNED_INT, pointer)
        self.after_draw(program)

    def calc_indices(self, num_vertices, vertices_per_face):
        num_faces = int(num_vertices / vertices_per_face)
        indices = []
        for i in range(num_faces):
            i0 = i * vertices_per_face
            ia = i0 + 1
            for _ in range(vertices_per_face - 2):
                ib = ia + 1
                indices.append(i0)
                indices.append(ia)
                indices.append(ib)
                ia = ib
        return indices

    def calc_normals(self, vertices):
        def normal(triangle):
            a, b, c = triangle
            v1 = b - a
            v2 = c - a
            n = v1 % v2

            direction = (a + b + c) / 3.0 - self.midpoint
            if n * direction < 0:
                n = -n

            return n

        normals = []
        for i in range(0, self.num_vertices, 3):
            n = normal(vertices[i:i+3])
            normals.extend([n, n, n])
        return normals

class WavefrontObject(Object):
    def __init__(self, obj):
        vertices, normals, indices, material_sequence = WavefrontObject.calc_arrays(obj)
        super(WavefrontObject, self).__init__(
            vertices=vertices,
            normals=normals,
            indices=indices,
            material=None,
        )
        self.material_sequence = material_sequence

    def draw(self, program):
        self.before_draw(program)

        if self.material_sequence:
            for j in range(len(self.material_sequence) - 1):
                i, m = self.material_sequence[j]
                next_i, _ = self.material_sequence[j + 1]
                program.use_material(m)
                glDrawElements(GL_TRIANGLES, next_i - i, GL_UNSIGNED_INT, c_void_p(sizeof(c_uint) * i));
            i, m = self.material_sequence[-1]
            program.use_material(m)
            glDrawElements(GL_TRIANGLES, self.num_indices - i, GL_UNSIGNED_INT, c_void_p(sizeof(c_uint) * i));

        self.after_draw(program)

    @staticmethod
    def calc_arrays(obj):
        vertices = []
        normals = []
        indices = []
        # stores tuples with the material and the starting index for the
        # elements vertices
        material_sequence = []

        # mapped with vertex index and normal index
        indices_dict = {}

        current_mtl = None
        material_map = {}
        def mtl_to_material(mtl):
            if mtl.name not in material_map:
                m = Material(
                    ambient=Vector3(*mtl.Ka),
                    diffuse=Vector3(*mtl.Kd),
                    specular=Vector3(*mtl.Ks),
                    specular_exponent=mtl.Ns,
                )
                material_map[mtl.name] = m
            return material_map[mtl.name]

        def index(i, j):
            if (i, j) not in indices_dict:
                indices_dict[(i, j)] = len(vertices)
                vertices.append(obj.vertices[i - 1][:3])
                normals.append(obj.normals[j - 1])
            return indices_dict[(i, j)]

        for f in obj.faces:
            vertex_data, mtl = f

            if not current_mtl or mtl.name != current_mtl.name:
                current_mtl = mtl
                material_sequence.append((len(indices), mtl_to_material(mtl)))

            i0 = index(vertex_data[0][0], vertex_data[0][2])
            ia = index(vertex_data[1][0], vertex_data[1][2])
            for i in range(2, len(vertex_data)):
                ib = index(vertex_data[i][0], vertex_data[i][2])
                indices.append(i0)
                indices.append(ia)
                indices.append(ib)
                ia = ib

        return vertices, normals, indices, material_sequence

class Program:
    shader_type_names = {
        GL_COMPUTE_SHADER: 'compute shader',
        GL_VERTEX_SHADER: 'vertex shader',
        GL_TESS_CONTROL_SHADER: 'tessellation control shader',
        GL_TESS_EVALUATION_SHADER: 'tessellation evaluation shader',
        GL_GEOMETRY_SHADER: 'geometry shader',
        GL_FRAGMENT_SHADER: 'fragment shader',
    }

    default_shaders = {
        GL_VERTEX_SHADER: '''
            #version 330 core

            layout (location = 0) in vec3 v;
            layout (location = 1) in vec3 normal;

            out vec3 v_normal;
            out vec3 v_v;

            uniform mat4 local = mat4(
                1.0, 0.0, 0.0, 0.0,
                0.0, 1.0, 0.0, 0.0,
                0.0, 0.0, 1.0, 0.0,
                0.0, 0.0, 0.0, 1.0
            );

            uniform mat4 model = mat4(
                1.0, 0.0, 0.0, 0.0,
                0.0, 1.0, 0.0, 0.0,
                0.0, 0.0, 1.0, 0.0,
                0.0, 0.0, 0.0, 1.0
            );

            uniform mat4 view = mat4(
                1.0, 0.0, 0.0, 0.0,
                0.0, 1.0, 0.0, 0.0,
                0.0, 0.0, 1.0, 0.0,
                0.0, 0.0, 0.0, 1.0
            );

            uniform mat4 proj = mat4(
                1.0, 0.0, 0.0, 0.0,
                0.0, 1.0, 0.0, 0.0,
                0.0, 0.0, 1.0, 0.0,
                0.0, 0.0, 0.0, 1.0
            );


            void main()
            {
                mat4 transform = model * local;
                v_v = vec3(transform * vec4(v, 1.0));
                v_normal = vec3(transform * vec4(normal, 1.0));

                gl_Position = proj * view * vec4(v_v, 1.0);
            }
        ''',
        GL_FRAGMENT_SHADER: '''
            #version 330 core

            in vec3 v_normal;
            in vec3 v_v;

            out vec4 color;

            struct Light {
                vec3 pos;
                vec3 ambient;
                vec3 diffuse;
                vec3 specular;
                float att_linear;
                float att_quad;
            };

            struct Material {
                vec3 ambient;
                vec3 diffuse;
                vec3 specular;
                float specular_exponent;
                float alpha;
            };

            uniform Light light = Light(
                vec3(1.0, 1.0, 1.0),
                vec3(1.0, 1.0, 1.0),
                vec3(1.0, 1.0, 1.0),
                vec3(1.0, 1.0, 1.0),
                0.0,
                0.0
            );

            uniform Material material = Material(
                vec3(.5, .5, .5),
                vec3(.5, .5, .5),
                vec3(.5, .5, .5),
                32.0,
                1.0
            );

            uniform vec3 camera_position = vec3(-1.0, 0.0, 0.0);

            void main()
            {
                vec3 view_direction = normalize(camera_position - v_v);
                vec3 light_direction = normalize(light.pos - v_v);

                vec3 normal = normalize(v_normal);
                /* If triangle is opposed to the camera, then color the other
                 * side. */
                if (dot(normal, view_direction) < 0)
                    normal = -normal;


                float cos_theta = dot(normal, light_direction);
                if (cos_theta < 0) {
                    cos_theta = 0;
                }
                vec3 diffuse = cos_theta * light.diffuse * material.diffuse;

                vec3 reflection = reflect(-light_direction, normal);
                cos_theta = dot(reflection, view_direction);
                if (cos_theta < 0) {
                    cos_theta = 0;
                }
                vec3 specular = pow(cos_theta, material.specular_exponent) *
                                light.specular * material.specular;

                vec3 ambient = light.ambient * material.ambient;

                float dist = length(light.pos - v_v);
                float attenuation = 1.0 / (1.0 + light.att_linear * dist + light.att_quad * dist * dist);

                color = vec4(attenuation * (diffuse + specular + ambient), material.alpha);
            }
        ''',
    }

    def __init__(self):
        self.shaders = dict(self.default_shaders)
        self.uniforms = {k: None for k in (
            'local',
            'model',
            'view',
            'proj',
            'camera_position',
            'light.pos',
            'light.ambient',
            'light.diffuse',
            'light.specular',
            'light.att_linear',
            'light.att_quad',
            'material.ambient',
            'material.diffuse',
            'material.specular',
            'material.specular_exponent',
            'material.alpha',
        )}

    def shader_code(self, shader_type, code):
        self.shaders[shader_type] = code

    def compile_and_link(self):
        self.program_id = glCreateProgram()
        if not self.program_id:
            raise Exception('Error on creating the OpenGL program')

        shader_ids = []
        try:
            for shader_type, code in self.shaders.items():
                shader_id = glCreateShader(shader_type)
                name = self.shader_type_names[shader_type]
                if not shader_id:
                    raise Exception("Error on creating %s" % name)
                shader_ids.append(shader_id)
                glShaderSource(shader_id, code)
                glCompileShader(shader_id)
                if glGetShaderiv(shader_id, GL_COMPILE_STATUS) != GL_TRUE:
                    error_string = glGetShaderInfoLog(shader_id)
                    raise Exception("Error on compiling %s: %s" % (name, error_string))
                glAttachShader(self.program_id, shader_id)
            glLinkProgram(self.program_id)
            if glGetProgramiv(self.program_id, GL_LINK_STATUS) != GL_TRUE:
                raise Exception('Error on linking the OpenGL program')
        finally:
            for i in shader_ids:
                glDeleteShader(i)

        for k in self.uniforms:
            self.uniforms[k] = glGetUniformLocation(self.program_id, k)
            if self.uniforms[k] < 0:
                raise Exception("Couldn't get location for uniform %s" % k)

    def use_material(self, material, enable_alpha=False):
        t = Vector3_to_tuple
        glUseProgram(self.program_id)
        glUniform3f(self.uniforms['material.ambient'], *t(material.ambient))
        glUniform3f(self.uniforms['material.diffuse'], *t(material.diffuse))
        glUniform3f(self.uniforms['material.specular'], *t(material.specular))
        glUniform1f(self.uniforms['material.specular_exponent'], material.specular_exponent)
        alpha = material.alpha if enable_alpha else 1.0
        glUniform1f(self.uniforms['material.alpha'], alpha)

    def use_light(self, light):
        t = Vector3_to_tuple
        glUseProgram(self.program_id)
        glUniform3f(self.uniforms['light.pos'], *t(light.position))
        glUniform3f(self.uniforms['light.ambient'], *t(light.ambient))
        glUniform3f(self.uniforms['light.diffuse'], *t(light.diffuse))
        glUniform3f(self.uniforms['light.specular'], *t(light.specular))
        glUniform1f(self.uniforms['light.att_linear'], light.att_linear)
        glUniform1f(self.uniforms['light.att_quad'], light.att_quad)

    def use_local_transform(self, transform):
        glUseProgram(self.program_id)
        m = transform.mat4()
        glUniformMatrix4fv(self.uniforms['local'], 1, GL_FALSE, m)

    def use_model_transform(self, transform):
        glUseProgram(self.program_id)
        m = transform.mat4()
        glUniformMatrix4fv(self.uniforms['model'], 1, GL_FALSE, m)

    def use_camera(self, camera):
        t = Vector3_to_tuple
        glUseProgram(self.program_id)
        m = camera.view_mat4()
        glUniform3f(self.uniforms['camera_position'], *t(camera.position))
        glUniformMatrix4fv(self.uniforms['view'], 1, GL_FALSE, m)

    def use_projection(self, projection):
        glUseProgram(self.program_id)
        m = projection.proj_mat4()
        glUniformMatrix4fv(self.uniforms['proj'], 1, GL_FALSE, m)




#!/usr/bin/env python

from __future__ import print_function
import os, wx, pickle
from MAVProxy.modules.lib import mp_util

'''
handle saving/loading of window positions
'''

window_list = {}
display_size = None
loaded_layout = None
pending_load = False
vehiclename = None

class WinLayout(object):
    '''represent window layout'''
    def __init__(self, name, pos, size, dsize):
        self.name = name
        self.pos = pos
        self.size = size
        self.dsize = dsize

    def __str__(self):
        return "%s(%ux%u@%u-%u)" % (self.name,
                                    self.size[0], self.size[1], 
                                    self.pos[0], self.pos[1])
        
class ManagedWindow(object):
    '''a layout plus callback for setting window position and size'''
    def __init__(self, layout, callback):
        self.layout = layout
        self.callback = callback

def get_wx_window_layout(wx_window):
    '''get a WinLayout for a wx window'''
    dsize = wx.DisplaySize()
    pos = wx_window.GetPosition()
    size = wx_window.GetSize()
    name = wx_window.GetTitle()
    return WinLayout(name, pos, size, dsize)

def set_wx_window_layout(wx_window, layout):
    '''set a WinLayout for a wx window'''
    try:
        wx_window.SetSize(layout.size)
        wx_window.SetPosition(layout.pos)
    except Exception as ex:
        print(ex)

def set_layout(wlayout, callback):
    '''set window layout'''
    global display_size
    global window_list
    global loaded_layout
    global pending_load
    global vehiclename
    #if not wlayout.name in window_list:
    #    print("layout %s" % wlayout)
    if not wlayout.name in window_list and loaded_layout is not None and wlayout.name in loaded_layout:
        callback(loaded_layout[wlayout.name])
    window_list[wlayout.name] = ManagedWindow(wlayout, callback)
    display_size = wlayout.dsize
    if pending_load:
        pending_load = False
        load_layout(vehiclename)

def layout_filename(fallback):
    '''get location of layout file'''
    global display_size
    global vehiclename
    (dw,dh) = display_size
    dirname = mp_util.dot_mavproxy()
    if vehiclename:
        fname = os.path.join(dirname, "layout-%s-%ux%u" % (vehiclename,dw,dh))
        if not fallback or os.path.exists(fname):
            return fname
    return os.path.join(dirname, "layout-%ux%u" % (dw,dh))

def save_layout(vehname):
    '''save window layout'''
    global display_size
    global window_list
    global vehiclename
    if display_size is None:
        print("No layouts to save")
        return
    vehiclename = vehname
    fname = layout_filename(False)
    if fname is None:
        print("No file to save layout to")
        return
    layout = {}
    try:
        # include previous layout, so we retain layouts for widows not
        # currently displayed
        layout = pickle.load(open(fname,"rb"))
    except Exception:
        pass
    count = 0
    for name in window_list:
        layout[name] = window_list[name].layout
        count += 1
    pickle.dump(layout, open(fname,"wb"))
    print("Saved layout for %u windows" % count)

def load_layout(vehname):
    '''load window layout'''
    global display_size
    global window_list
    global loaded_layout
    global pending_load
    global vehiclename
    if display_size is None:
        pending_load = True
        return
    vehiclename = vehname
    fname = layout_filename(True)
    if fname is None:
        print("No file to load layout from")
        return
    try:
        layout = pickle.load(open(fname,"rb"))
    except Exception:
        layout = {}
        print("Unable to load %s" % fname)
        loaded_layout = layout
        return
    count = 0
    for name in window_list:
        if name in layout:
            try:
                window_list[name].callback(layout[name])
                count += 1
            except Exception as ex:
                print(ex)
    loaded_layout = layout
    print("Loaded layout for %u windows" % count)
    
